package fi.helsinki.dacopan;

import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.prefs.Preferences;

import javax.swing.JMenuItem;

/**
 * Localized message provider.
 * 
 * @version $Id: Localization.java,v 1.12 2004/05/22 12:01:12 vainio Exp $
 */
public class Localization {

    /** name of the resource bundle */
    private static final String BUNDLE_NAME = "fi.helsinki.dacopan.dacopan";
    
    /** classpath resource for the list of available locales */
    private static final String LOCALE_RESOURCE = "locales.properties";
    
    /** key under which language selection is stored in preferences */
    private static final String PREF_LANGUAGE_KEY = "dacopan.ui.language";
    
    private static ResourceBundle resourceBundle;
    private static Language currentLanguage;
    private static Language defaultLanguage;
    private static Map availableLanguages;

    //Static initialization block
    static {
        initLocales();
    }
    
    private Localization() {
    }
    
    /**
     * Initializes the locale list from classpath.
     */
    private static void initLocales() {

        // locate the resource
        InputStream is = Localization.class.getResourceAsStream(LOCALE_RESOURCE);
        if(is == null) {
            throw new RuntimeException("locale.properties not found on classpath");
        }
        Properties p = new Properties();
        try {
            p.load(is);
        } catch (IOException e) {
            throw new RuntimeException("Couldn't read locale.properties", e);
        }
        
        // read in the locale information
        String defaultLocaleCode = p.getProperty("default");
        availableLanguages = new LinkedHashMap();
        String[] localeCodes = p.getProperty("locales").split("\\s*,\\s*");
        for (int i = 0; i < localeCodes.length; i++) {
            Locale loc = new Locale(localeCodes[i]);
            Language lang = new Language(p.getProperty("locale." + localeCodes[i]), loc);
            if(localeCodes[i].equals(defaultLocaleCode)) {
                defaultLanguage = lang;
            }
            availableLanguages.put(loc.getLanguage(), lang);
        }
        
        // if predefined language, use it
        Preferences prefs = Preferences.userNodeForPackage(Localization.class);
        String preferredLocaleCode = prefs.get(PREF_LANGUAGE_KEY, null);
        if(preferredLocaleCode != null) {
            if((currentLanguage = (Language)availableLanguages.get(preferredLocaleCode)) == null) {
                System.out.println("User's preferred language (" + preferredLocaleCode + ") was NOT found");
            } else {
                System.out.println("Using user's preferred language, " + currentLanguage.getLocale());
            }
        }
        
        // if none selected yet, try user's default
        if(currentLanguage == null) {
            String userDefault = Locale.getDefault().getLanguage();
            if((currentLanguage = (Language)availableLanguages.get(userDefault)) == null) {
                System.out.println("No localization for system default language, " + userDefault);
            } else {
                System.out.println("Using system default language, " + currentLanguage.getLocale());
            }
        }
        
        // if all else fails, use default locale of ours
        if(currentLanguage == null) {
            currentLanguage = defaultLanguage;
            System.out.println("No language selected, using our default, " + currentLanguage.getLocale());
        }
        
        Locale.setDefault(defaultLanguage.getLocale());
        setLanguage(currentLanguage);
        
    }
    
    /**
     * Sets the language of the Animator, gets the correct bundle
     * for the locale.
     */
    private static void setLanguage(Language language) {
        currentLanguage = language;
        resourceBundle = ResourceBundle.getBundle(BUNDLE_NAME, language.getLocale());
        System.out.println("Language set to " + language);
    }
    
    /**
     * Gets the current locale.
     */
    public static Locale getCurrentLocale() {
        return currentLanguage.getLocale();
    }
    
    /**
     * Gets the current language.
     */
    public static Language getCurrentLanguage() {
        return currentLanguage;
    }

    /**
     * Sets the language to be used in the whole animator user interface.
     */
    public static void setCurrentLanguage(Language language) {
        if(language != null) {
            setLanguage(language);
            Preferences prefs = Preferences.userNodeForPackage(Localization.class);
            prefs.put(PREF_LANGUAGE_KEY, language.getLocale().getLanguage());
        }
    }

    /**
     * Gets a collection of available languages.
     */
    public static Collection getAvailableLanguages() {
        return availableLanguages.values();
    }

    /**
     * Gets a localized string for the given key.
     * 
     * @param key for lookup
     * @return localized string
     */
    public static String getString(String key) {
        return getStringForKey(key);
    }

    /**
     * Convenience shortcut to {@link #getString(String, Object[])}.
     */
    public static String getString(String key, Object a) {
        return getString(key, new Object[] { a });
    }
    
    /**
     * Convenience shortcut to {@link #getString(String, Object[])}.
     */
    public static String getString(String key, Object a, Object b) {
        return getString(key, new Object[] { a, b });
    }
    
    /**
     * Gets a localized string populated with the given parameters.
     */
    public static String getString(String key, Object[] params) {
        return new MessageFormat(getStringForKey(key), currentLanguage.getLocale())
            .format(params, new StringBuffer(), null).toString();
    }
    
    /**
     * Localizes the text and (possible) mnemonic of a menu item.
     */
    public static void localizeMenuItem(JMenuItem item, String key) {
        String value = getStringForKey(key);
        String text = value.replaceAll("\\$", "");
        item.setText(text);
        int pos = value.indexOf('$');
        if(pos != -1 && pos + 1 < value.length()) {
            item.setMnemonic(value.charAt(pos + 1));
            item.setDisplayedMnemonicIndex(pos);
        }
    }
    
    private static String getStringForKey(String key) {
        try {
            return resourceBundle.getString(key);
        } catch (MissingResourceException e) {
            System.out.println("ERROR: No value for localization key " + key);
            return '!' + key + '!';
        }
    }

    /**
     * A wrapper for a locale and the (native) name of its language.
     */
    public static final class Language {
        private String name;
        private Locale locale;
        
        private Language(String name, Locale locale) {
            this.name = name;
            this.locale = locale;
        }
        
        public Locale getLocale() {
            return locale;
        }
        
        public String getName() {
            return name;
        }
        
        public String toString() {
            return name + "/" + locale;
        }
        
        public boolean equals(Object obj) {
            if(obj instanceof Language) {
                return ((Language)obj).getLocale().equals(locale);
            }
            return false;
        }
    }
    
}