/*
 * AnimationSequence.java
 */

package fi.helsinki.dacopan.animseq;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JOptionPane;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.scenario.Saveable;
import fi.helsinki.dacopan.ui.MainFrame;
import fi.helsinki.dacopan.ui.SettingsMSC;
import java.awt.Dimension;
import java.awt.Point;

/**
 * A class to handle the logic of the animation sequence (also called scenario play list). This
 * class encapsulates the list itself and all direct references to list items. This class
 * provides functionality for recording items, deleting items, playing the list in sequential
 * order and for navigating the list by showing each individual item item upon request.
 * 
 * <p>The play list consists of specific item. Each item is either showable or non-showable.
 * Non-showable items are some kind of markers, and selecting them from the list causes no actions.
 * The end marker at the end of the list is one such marker. All showable items, when selected, 
 * cause the UI of the application to present some type of visualisation.
 *
 * <p>Inserting or recording new items to the list is possible when the recording mode is on. In
 * that case some user actions (such as pressing play, changing the layer, or showing the
 * encapsulation for a unit) are automatically interpreted (by MainFrame) as recording actions.
 * Other recording actions are manually initiated by selecting buttons in the
 * ScenarioEditorPanel. New items are always inserted in the insertion positions (in the place
 * of the current item) in the list. At any time, at most one 'unfinished' item may exist in the
 * list. An unfinished item is an MSC item with the start time recorded but the end time missing.
 * In cases of new recording actions while there still is an unfinished item, the
 * AnimationSequence tries to finish the unfinished item by recording the 'now time' as the
 * end time for the unfinished item.
 *
 * <p>The actions of scenario play list are closely related to the scenario ScenarioEditorDialog.
 * Such a dialog should always be is existence. The dialog can be hidden, but never closed and
 * disposed of.
 *
 * <p>Some methods in this class are purposefully package private, so that they can only be called
 * by classes in the same package. This is done to enforce encapsulation of the logic inside this
 * package.
 */
public class AnimationSequence 
    implements Saveable
{
    
    private static final boolean PLAY_AFTER_SHOWING_ENC = false;
    
    private MainFrame mainFrame;

    /** A list containing ScenarioItem objects. The last item on the list should ALWAYS be
     * ScenarioItemEndMarker.
     */
    private List sequence;
    
    private ScenarioEditorDialog scenarioDialog;
    private ScenarioEditorPanel scenarioPanel;
    
    /** When true, AnimationSequence is in recording mode. All recording needs to happen in
     * recording mode.
     */
    private boolean isRecording = false;

    /** True when changing from one item to another. Used to prevent circular event handling. */
    private boolean isChangingMode = false;
    
    /** True when showing an Enc item from animation sequence, false otherwise. Particularly,
     * this will be false when Animator is showing an Enc that the user has chosen manually without
     * recording it in animation sequence. When true, coming out of Enc mode (in MainFrame) causes
     * a step forward in animation sequence. This is not done when isShowingEncFromList is false.
     */
    private boolean isShowingEncFromList = false;
        
    /** References an item (currently ALWAYS ScenarioItemMSC) that has start time recorded but
     * not end time.
     */
    private ScenarioItem unfinishedItem = null;
    
    /** The index of the list item currently shown. */
    private int indexShown = 0;
    
    /**
     * Creates a new instance of AnimationSequence
     */
    public AnimationSequence(MainFrame frame) {
        this.mainFrame = frame;
        this.sequence = new LinkedList();
        this.sequence.add(new ScenarioItemEndMarker());
        this.scenarioDialog = new ScenarioEditorDialog(mainFrame, this);
        this.scenarioPanel = scenarioDialog.getScenarioPanel();
    }
    
    /** Moves to the next item in play list and executes the action associated with it. Does
     * nothing when called in record mode, or when the list is already at the end.
     *
     * @param playNext If true, the AnimationTimeState is set to play mode after executing the
     * action within the item.
     */
    public void showNext() {
        if (indexShown < sequence.size()-1) {
            indexShown++;
            doActionForIndex(indexShown);
        }
    }
    
    void showIndex(int index) {
        if (index >= 0 && index < sequence.size()) {
            indexShown = index;
            doActionForIndex(indexShown);
        }
    }
    
    private void doActionForIndex(int index) {
        ScenarioItem item = (ScenarioItem)sequence.get(index);
        if (item instanceof ScenarioItemENC) {
            isShowingEncFromList = true;
        } else {
            isShowingEncFromList = false;
        }
        isChangingMode = true;
        item.doAction(mainFrame, isRecording);
        isChangingMode = false;
        scenarioPanel.refreshWidgets();
    }
    
    
    /** Either shows the scenario play list dialog or hides it.
     *
     * @param isScenarioModeActive If true, the dialog is shown, if false, the dialog is hidden.
     */    
    public void setActive(boolean isScenarioModeActive) {        
        if (isScenarioModeActive) {
            scenarioDialog.show();
            scenarioPanel.refreshWidgets();
        } else {
            scenarioDialog.hide();
        }
    }
    
    /** The MainFrame should call this method when quitting the enc mode. That way the next item
     * in the list can be properly displayed. This method can always be called when coming out
     * of enc mode, the AnimationSequence is responsible for ignoring unnecessary calls.
     */
    public void notifyExitFromEnc() {
        if (isShowingEncFromList && indexShown < sequence.size()-1) {
            indexShown++;
            doActionForIndex(indexShown);
            if (PLAY_AFTER_SHOWING_ENC) {
                mainFrame.getAnimationTimeState().play();
            }
            isShowingEncFromList = false;
        } else {
            scenarioPanel.refreshWidgets();
        }
    }
    
    // package private so that only the scenarioeditor stuff in the same package can call this
    void setRecordingMode(boolean newMode) {
        this.isRecording = newMode;
    }
    
    /** Returns the visible data for the scenario play list.
     *
     * @return A String[] array with each element representing one line in the play list.
     */    
    public String[] getPlaylist() {
        String[] playlist = new String[sequence.size()];
        int index = 0;
        Iterator iterator = sequence.iterator();
        while (iterator.hasNext()) {
            ScenarioItem element = (ScenarioItem)iterator.next();
            playlist[index++] = element.getPlaylistText();
        }
        
        return playlist;
    }
    
    /** Should be called when the start time for a new MSC item should be recorded. If the 
     * recording mode is not on, calling this does nothing.
     *
     * @param settings The settings for the new item.
     * @param startTime The start time for the new item.
     */    
    public void recordStartMSC(SettingsMSC settings, float startTime) {
        if (isRecording) {
            if (!finishUnfinished()) {
                return;
            }
            unfinishedItem = new ScenarioItemMSC(settings, startTime);
            addSequenceItem(unfinishedItem);
        }
    }
    
    void recordEndSelected() {
        if (isRecording && (unfinishedItem != null)) {
            float nowTime = mainFrame.getAnimationTimeState().getNowTime();
            boolean success = unfinishedItem.recordEnd(nowTime);
            if (success) {
                unfinishedItem = null;
                scenarioPanel.refreshWidgets();
            } else {
                JOptionPane.showMessageDialog(scenarioDialog, 
                    Localization.getString("animseq.end_time_before_start"),  //$NON-NLS-1$
                    Localization.getString("animseq.recording_error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
            }
        }
    }
    
    void recordStartSelected() {
        if (isRecording && (unfinishedItem == null)) {
            float nowTime = mainFrame.getAnimationTimeState().getNowTime();
            SettingsMSC settings = 
                mainFrame.getSettings().getSettingsMSC(mainFrame.getCurrentLayer());
            recordStartMSC(settings, nowTime);
        }
    }
    
    void recordEncSelected() {
        if (isRecording && (unfinishedItem == null)) {
            TransferUnit unit = mainFrame.getCurrentENCUnit();
            recordEnc(unit);
        }
    }
    
    /** Instructs the AnimationSequence to record a new ENC item in the current insertion
     * position. A new item will only be recorded if the recording mode is on.
     *
     * @param unit The selected unit in the enc diagram.
     */    
    public void recordEnc(TransferUnit unit) {
        if (isChangingMode) return;
        
        if (isRecording && (unit != null)) {
            // when going to Enc, record end for unfinished item
            if (!finishUnfinished()) {
                return;
            }

            addSequenceItem(new ScenarioItemENC(unit));
        } else {
            // if not recording, then just refresh widgets
            scenarioPanel.refreshWidgets();
        }
    }
    
    private boolean finishUnfinished() {
        if (unfinishedItem != null) {
            float nowTime = mainFrame.getAnimationTimeState().getNowTime();
            boolean success = unfinishedItem.recordEnd(nowTime);
            if (success) {
                unfinishedItem = null;
                return true;
            } else { 
                JOptionPane.showMessageDialog(scenarioDialog, 
                    Localization.getString("animseq.unfinished_item"),  //$NON-NLS-1$
                    Localization.getString("animseq.recording_error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
                scenarioPanel.refreshWidgets();
                return false;
            }
        } else {
            return true;
        }
    }
    
    private void addSequenceItem(ScenarioItem item) {
        if (item == null) {
            throw new IllegalArgumentException("Cannot add a null item!");
        }
        
        sequence.add(indexShown, item);
        indexShown++;
        scenarioPanel.refreshWidgets();
        mainFrame.setFileModified(true);
    }
    
    void deleteItemSelected() {
        if (indexShown != sequence.size()-1) {
            Object removedItem = sequence.remove(indexShown);
            if (unfinishedItem == removedItem) {
                unfinishedItem = null;
            }
            doActionForIndex(indexShown);
            mainFrame.setFileModified(true);
        }
    }
    
    void editItemSelected() {
        if (canEditCurrentItem()) {
            mainFrame.invokeChangeSettings();
        }
    }
    
    private boolean canEditCurrentItem() {
        return (sequence.get(indexShown) instanceof ScenarioItemMSC);
    }
    
    void recordPauseSelected() {
        if (!finishUnfinished()) {
            return;
        }
        
        addSequenceItem(new ScenarioItemPause());
    }
    
    void nextItemSelected() {
        showNext();
    }
    
    boolean isRecordEndPossible() {
        return (isRecording && mainFrame.isInMSCMode() && (unfinishedItem != null));
    }
    
    boolean isRecordStartPossible() {
        return (isRecording && mainFrame.isInMSCMode() && (unfinishedItem == null));
    }
    
    boolean isRecordEncPossible() {
        return isRecording && mainFrame.isInENCMode();
    }
    
    boolean isRecordPausePossible() {
        return isRecording;
    }
    
    boolean isNextItemPossible() {
        return (indexShown < sequence.size()-1);
    }
    
    boolean isEditItemPossible() {
        return isRecording && canEditCurrentItem();
    }
    
    int getIndexShown() {
        return indexShown;
    }
    
    /** Returns the list data for saving in a scenario file.
     *
     * @return The scenario play list.
     */    
    public Object getData() {
        return sequence;
    }
    
    /** Initialises the state of the AnimationTimeState and replaces the old play list with
     * the new list given as parameter.
     *
     * @param o New scenario list, as loaded from a scenario file.
     */    
    public void setData(Object o) {
        this.sequence = (List)o;
        this.isRecording = false;
        this.isChangingMode = false;
        isShowingEncFromList = false;
        unfinishedItem = null;
        indexShown = 0;
    }
    
    /**
     * Returns true if animation sequence contains showable items (items that will set MainFrame
     * to some state) and false otherwise.
     * @return
     */
    public boolean hasShowables() {
        Iterator iterator = sequence.iterator();
        while (iterator.hasNext()) {
            ScenarioItem element = (ScenarioItem)iterator.next();
            if (element.getShowable()) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Shows the first showable item. If showable items are available, the first one is
     * shown and true is recorded. If no showables are available, calling this method does
     * nothing, and false is returned.
     * @return
     */
    public boolean showFirstShowable() {
        Iterator iterator = sequence.iterator();
        int tempIndex = 0;
        while (iterator.hasNext()) {
            ScenarioItem element = (ScenarioItem)iterator.next();
            if (element.getShowable()) {
                showIndex(tempIndex);
                return true;
            }
            tempIndex++;
        }
        
        return false;
    }
    
    /** Completely re-creates the scenario window with all its contents. Should be used e.g. when
     * the locale is changed, to refresh the window UI to match the new locale. Calling this
     * method preserves the location, size and visibility status of the old window.
     */
    public void refreshWindow() {
        Point oldLocation = scenarioDialog.getLocation();
        Dimension oldSize = scenarioDialog.getSize();
        boolean previouslyVisible = scenarioDialog.isVisible();
        
        scenarioDialog.dispose();
        
        scenarioDialog = new ScenarioEditorDialog(mainFrame, this);
        scenarioPanel = scenarioDialog.getScenarioPanel();
        
        scenarioDialog.setLocation(oldLocation);
        scenarioDialog.setSize(oldSize);
        
        scenarioDialog.setVisible(previouslyVisible);
        scenarioPanel.refreshWidgets();
    }

	/** Returns true if the scenario dialog is visible, false otherwise. */
	public boolean isActive() { 
		return scenarioDialog.isVisible();
	}
    
    /** Returns true if AnimationSequence is in recording mode, false otherwise. */
    public boolean isInRecordingMode() {
        return isRecording;
    }
    
    /** This method should be called from MainFrame whenever the user tries to change layer. In
     * the case that recording a scenario is in such a state that changing the layer would not
     * be a valid action, false is returned. If false is returned, the layer MUST NOT be changed.
     *
     * @return True if layer can be changed, false otherwise.
     */    
    public boolean layerChanged() {
        if (finishUnfinished()) {
            scenarioPanel.refreshWidgets();
            return true;
        } else {
            return false;
        }
    }
    
}
