/*
 * AnimationTimeState.java
 *
 * Created on 9. huhtikuuta 2004, 21:55
 */

package fi.helsinki.dacopan.contsig;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.Timer;

import fi.helsinki.dacopan.animseq.AnimationSequence;
import fi.helsinki.dacopan.model.StepIterator;

/**
 * 
 */
public class AnimationTimeState implements java.awt.event.ActionListener {
    
    private Timer timer;
    private List listeners;
    private AnimationSequence animationSequence;
    
    private float nowTime;
    
    /** When the toEnd call is received, the nowTime is moved to animationEndTime. Thus,
     * animationEndTime represents the end of available data.
     */
    private float animationEndTime;
    
    /** If the animationSequence is configured for this AnimationTimeState object, the
     * AnimationSequence.showNext() is called when the scenarioItemEndTime is reached in
     * play mode (reaching it by stepping does not count). Thus, the scenarioItemEndTime
     * represents the end time for a scenario item in scenario mode.
     */
    private float scenarioItemEndTime = -1;
    private float timeScale;
    
    /** The same as refresh delay. */
    private int stepInterval;
    
    private float stepSize;
    private boolean isInPlayMode;
    
    private StepIterator stepIterator;
    private boolean previouslyStepping;
    private long previousPlayEvent = -1;
    

    /** Creates a new instance of AnimationTimeState */    
    public AnimationTimeState(
        float nowTime, 
        float animationEndTime, 
        float timeScale,
        int stepInterval,
        StepIterator stepIterator) 
    {
        if (nowTime < 0f) {
            throw new IllegalArgumentException("nowTime cannot be negative!");
        }
        
        if (timeScale <= 0f) {
            throw new IllegalArgumentException("timeScale cannot be negative or zero!");
        }
        
        if (stepInterval <= 0) {
            throw new IllegalArgumentException("stepInterval cannot be negative or zero!");
        }
        
        if (animationEndTime < nowTime) {
            throw new IllegalArgumentException("endTime cannot be before nowTime!");
        }
        
        if (stepIterator == null) {
            throw new IllegalArgumentException("stepIterator cannot be null!");
        }
        
        this.nowTime = nowTime;
        this.animationEndTime = animationEndTime;
        this.timeScale = timeScale;
        this.stepInterval = stepInterval;
        
        this.isInPlayMode = false;
        this.previouslyStepping = false;
        this.stepSize = 1.0f / timeScale / (1000 / stepInterval);
        this.timer = new Timer(stepInterval, this);
        this.listeners = new LinkedList();
        this.animationSequence = null;

        this.stepIterator = stepIterator;
    }
    
    public AnimationTimeState(
        float nowTime, 
        float animationEndTime, 
        float timeScale,
        int stepInterval,
        StepIterator stepIterator,
        AnimationSequence animationSequence,
        float scenarioItemEndTime) 
    {
        this(nowTime, animationEndTime, timeScale, stepInterval, stepIterator);
        this.animationSequence = animationSequence;
        this.scenarioItemEndTime = scenarioItemEndTime;
    }
    
        
    
    // PRIVATE METHODS
    
    private void scenarioItemEndReached() {
        nowTime = scenarioItemEndTime;
        if (animationSequence != null) animationSequence.showNext();
    }
    
    private void animationEndReached() {
        nowTime = animationEndTime;
        pause();
    }

    
    private void stepListenersToNowTime() {
        Iterator iter = listeners.iterator();
        while (iter.hasNext()) {
            ControlSignalsListener element = (ControlSignalsListener)iter.next();
            element.stepTo(nowTime);
        }
    }        

    // INTERFACE IMPLEMENTATION
    
    public void actionPerformed(java.awt.event.ActionEvent event) {
        if (event.getSource() != timer) return;
        if (!isInPlayMode) return;
        
        long sysTime = System.currentTimeMillis();
        
        float localStepSize = this.stepSize;
        
        if (previousPlayEvent > 0) {
            int trueInterval = (int)(sysTime - previousPlayEvent);
            localStepSize *= (float)trueInterval / (float)stepInterval;
        }

        if (nowTime + localStepSize > animationEndTime) {
            localStepSize = animationEndTime - nowTime;
        }
        
        if ((scenarioItemEndTime >= 0f) && (nowTime + localStepSize > scenarioItemEndTime)) {
            localStepSize = scenarioItemEndTime - nowTime;
        }
        
        nowTime += localStepSize;
        
        Iterator iter = listeners.iterator();
        while (iter.hasNext()) {
            ControlSignalsListener element = (ControlSignalsListener)iter.next();
            element.advance(localStepSize, nowTime);
        }
        
        if ((scenarioItemEndTime >= 0f) && (nowTime >= scenarioItemEndTime)) {
            scenarioItemEndReached();
        }
        
        if (nowTime >= animationEndTime) {
            animationEndReached();
        }
        
        previouslyStepping = false;
        previousPlayEvent = sysTime;
    }

    // GETTERS & SETTERS
    
    /** 
     * Getter for property timeScale.
     * @return Value of property timeScale.
     */
    public float getTimeScale() {
        return timeScale;
    }    

    /** 
     * Setter for property timeScale.
     * @param timeScale New value of property timeScale.
     */
    public void setTimeScale(float timeScale) {
        this.timeScale = timeScale;
        this.stepSize = 1.0f / timeScale / stepInterval;
    }    
    
    /** 
     * Getter for property stepInterval.
     * @return Value of property stepInterval.
     */
    public int getStepInterval() {
        return stepInterval;
    }
    
    /** 
     * Setter for property stepInterval.
     * @param stepInterval New value of property stepInterval.
     */
    public void setStepInterval(int stepInterval) {
        this.stepInterval = stepInterval;
        this.stepSize = 1.0f / timeScale / stepInterval;
        this.timer.setDelay(stepInterval);
    }
    
    /** Gets the nowTime. */
    public float getNowTime() {
        return nowTime;
    }

    /** Sets the nowTime and steps all listeners to show that time. If the given
     * parameter is less than zero, time is set to zero. If the parameter is greater than
     * animationEndTime, time is set to animationEndTime.
     */
    public void setNowTime(float newNowTime) {
        previouslyStepping = false;
        
        if (newNowTime < 0f) newNowTime = 0f;
        else if (newNowTime > this.animationEndTime) newNowTime = this.animationEndTime;
        if (newNowTime != this.nowTime) {
            this.nowTime = newNowTime;
            stepListenersToNowTime();
        }
    }

    /**
     * @param animationSequence
     */
    /*public void setAnimationSequence(AnimationSequence animationSequence) {
        this.animationSequence = animationSequence;        
    }*/
    
    /** Returns true if the animation is currently paused, false otherwise. */
    public boolean isPaused() {
        return !isInPlayMode;
    }
    
    // LISTENER METHODS
    
    /** Adds a ControlSignalsListener as a listener for this AnimationTimeState. */
    public void addControlSignalsListener(ControlSignalsListener toAdd) {
        if (this.listeners.contains(toAdd)) return;
        this.listeners.add(toAdd);
    }
    
    /** Removes a ControlSignalsListener from the listener list, if it exists in the list. */
    public void removeControlSignalsListener(ControlSignalsListener toRemove) {
        if (this.listeners == null) return;
        this.listeners.remove(toRemove);
    }
    
    // TIME MANIPULATION METHODS
    
    /** Puts the AnimationTimeState in play mode. In play mode the AnimationTimeState will
     * continuously fire advance events (according to a timer) until the animation is played
     * the the end, or AnimationTimeState is put to pause mode.
     * 
     * <p>Switching to play mode will cause the AnimationTimeState to fire the toPlayMode event
     * once.
     */
    public void play() {
        if (nowTime == animationEndTime) return;
        this.previousPlayEvent = -1;
        this.isInPlayMode = true;
        this.timer.start();
        Iterator iter = listeners.iterator();
        while (iter.hasNext()) {
            ControlSignalsListener element = (ControlSignalsListener)iter.next();
            element.toPlayMode();
        }
    }
    
    /** Puts the AnimationTimeState in pause mode. In pause mode the AnimationTimeState reacts
     * to other calls (toBeginning, toEnd, stepForward, stepBackward), but does not fire events
     * according to a timer.
     *
     * <p>Switching to pause mode will cause the AnimationTimeState to fire the toPauseMode event
     * once.
     */
    public void pause() {
        this.previousPlayEvent = -1;
        this.isInPlayMode = false;
        this.timer.stop();
        Iterator iter = listeners.iterator();
        while (iter.hasNext()) {
            ControlSignalsListener element = (ControlSignalsListener)iter.next();
            element.toPauseMode();
        }
    }
    
    public void toBeginning() {
        this.previousPlayEvent = -1;
        this.nowTime = 0f;
        if (stepIterator != null) {
            stepIterator.first();
            previouslyStepping = true;
        }
        stepListenersToNowTime();
    }
    
    public void toEnd() {
        this.previousPlayEvent = -1;
        this.nowTime = animationEndTime;
        if (stepIterator != null) {
            stepIterator.last();
            previouslyStepping = true;
        }
        stepListenersToNowTime();
    }
    
    public void toEndOfScenarioItem() {
        if (this.scenarioItemEndTime >= 0f) {
            this.previousPlayEvent = -1;
            this.nowTime = this.scenarioItemEndTime;
            this.previouslyStepping = false;
            stepListenersToNowTime();
        }
    }
    
    public void stepForward() {
        this.previousPlayEvent = -1;
        if (stepIterator != null) {
            if (previouslyStepping) {
                if (stepIterator.hasNext()) {
                    nowTime = stepIterator.next();
                    stepListenersToNowTime();
                }
            } else {
                nowTime = stepIterator.getNextForTime(nowTime);
                previouslyStepping = true;
                stepListenersToNowTime();
            }
        } else {
            setNowTime(nowTime+stepSize);
        }
    }
    
    public void stepBackward() {
        this.previousPlayEvent = -1;
        if (stepIterator != null) {
            if (previouslyStepping) {
                if (stepIterator.hasPrevious()) {
                    nowTime = stepIterator.previous();
                    stepListenersToNowTime();
                }
            } else {
                nowTime = stepIterator.getPreviousForTime(nowTime);
                previouslyStepping = true;
                stepListenersToNowTime();
            }
        } else {
            setNowTime(nowTime-stepSize);
        }
    }

    /** Stops the timer without sending any more events. */
    public void discard() {
        timer.stop();
    }
    
    
    // METHODS FOR QUERYING THE CURRENT POSITION IN TIME
    
    /**
     * Tells whether the animation time state is at the beginning
     * of the current animation.
     */
    public boolean isAtBeginning() {
        if(nowTime == 0) {
        	return true;
        }
        else {
        	return false;
        }
    }
    
    /**
     * Tells whether the animation time state is at the end
     * of the current animation.
     */
    public boolean isAtEnd() {
    	if(nowTime == animationEndTime) {
    		return true;
        }
        else {
        	return false;
        }
    }
}
