package fi.helsinki.dacopan.model;

import java.util.List;

/**
 * Provides different views to the core data structures.
 * 
 * TODO consider using Collection instead of List for every method whose
 *      return value doesn't explicitly require a List
 * 
 * @version $Id: DataView.java,v 1.15 2004/05/11 09:21:05 aarniala-dacopan Exp $
 */
public interface DataView {
    
    /**
     * Adds the given host to this view.
     */
    void addHost(Host host);
    
    /**
     * Adds the given flow to this view.
     */
    void addFlow(Flow flow);
    
    /**
     * Adds the given link to this view.
     */
    void addLink(Link link);
    
    /**
     * Adds the given layer to this view.
     */
    void addLayer(Layer layer);
    
    /**
     * Adds the given protocol to this view.
     */
    void addProtocol(Protocol protocol);
    
    /**
     * Adds a transfer unit to the view. If <code>recurse</code> is
     * true, the method will recurse to the children of the unit and
     * add them to the view as well.
     * 
     * @param unit to add
     * @param recurse whether to recurse
     */
    void addTransferUnit(TransferUnit unit, boolean recurse);

    /**
     * Returns a sorted list of all the {@link TransferUnit units} in
     * the given layer.
     * 
     * @param layer whose units to return
     */
    List getUnitsForLayer(Layer layer);

    /**
     * Returns a collection of all the {@link TransferUnit units} in
     * the given layer between the given timestamps (inclusive).
     * 
     * @param layer whose units to return
     * @param start timestamps
     * @param end timestamps
     */
    List getUnitsForLayer(Layer layer, float start, float end);
    
    /**
     * Returns the transfer unit with the given id, if any.
     */
    TransferUnit getUnitById(long id);
    
    /**
     * Returns an {@link StepIterator iterator} for the steps on
     * the specified layer.
     * 
     * @param layer to step thru
     * @return iterator for stepping
     */
    StepIterator getStepIterator(Layer layer);
    
    /**
     * Returns the end time of the network exchange data, as the timestamp of the last event
     * included in the data, regardless of the type of event, layer etc.
     */
    float getEndTime();
    
    /**
     * Returns the end time of the network exchange data for a particular layer, as the timestamp 
     * of the last event in the layer.
     */
    float getEndTimeForLayer(Layer layer);
    
    /**
     * Returns the maximum length of all the values of the given variable.
     */
    int getValueMaxLength(VariableDefinition def);
    
    /**
     * Returns a list of all the available variables (those that have values
     * longer than zero characters) on the specified layer.
     */
    List getAvailableVariables(Layer layer);
    
    /**
     * Returns all the {@link Host hosts} present in the scenario.
     */
    List getHosts();
    
    /**
     * Returns the host by the id.
     */
    Host getHostById(long id);

    /**
     * Returns all the {@link Flow flows} present in the scenario.
     */
    List getFlows();
    
    /**
     * Returns the flow by the given id, if any.
     */
    Flow getFlowById(long id);
    
    /**
     * Returns all the layers present in the scenario, ordered from lower to higher.
     * (for instance Network -> Transport -> Application)
     */
    List getLayers();
    
    /**
     * Returns the layer with the given id.
     */
    Layer getLayerById(long id);
    
    /**
     * Returns all the protocols present.
     */
    List getProtocols();
    
    /**
     * Returns the protocol with the given id, if any.
     */
    Protocol getProtocolById(long id);

    /**
     * Returns a list of all the {@link Link links} in the scenario.
     */
    List getLinks();
    
    /**
     * Returns the link with the given id, if any.
     */
    Link getLinkById(long id);
    
}