package fi.helsinki.dacopan.model;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;

import fi.helsinki.dacopan.scenario.Saveable;

/**
 * This class acts as a container for all Notes. It takes care of providing an
 * ecapsulated notes management strategy, so that each component that need notes
 * data uses the services in this class.
 *
 * In the design this is sometimes referred to as the "notes interface".
 *
 */
public class NoteManager implements Saveable {
    
    /**
     * Notes attached to time and layer are stored in one map, with as many entries
     * as there are layers. The layer acts as the key, and the value is a LinkedList
     * of all the Notes on that layer, in time order.
     * Encapsulation notes that are attached to transfer units are stored
     * in a hash map with the TransferUnit as the key.
     */
    protected Map notesMsc;
    protected Map notesEnc;
    
    /**
     * This constructor sets up the proper HashMaps needed
     * for keeping track of the notes relevant to a
     * particular set of protocol events. (a scenario)
     *
     * @param dataView
     */
    public NoteManager() {
        //create an empty hashmap for the layer lists
        notesMsc = new HashMap();
        //and then we init a new hashmap for the ENC notes
        notesEnc = new HashMap();
    }
    /**
     * Adds a note to the requested layer by putting it in the corresponding
     * linked list.
     *
     * @param note
     * @return
     */
    public void addNoteTimeLayer(Note note) {
        LinkedList relevantLayerList = getNotesForLayer(note.getLayer());
        relevantLayerList.add(note);
        Collections.sort(relevantLayerList);
    }
    
    /**
     * Deletes the passed note from the NoteManager
     * @param note
     */
    public void deleteNote(Note note) throws IllegalArgumentException {
        if (note.getLayer() != null) {
            LinkedList relevantLayerList = getNotesForLayer(note.getLayer());
            if (!relevantLayerList.contains(note)) {
                throw new IllegalArgumentException("Note not found in its layer's linked list!");
            } else {
                relevantLayerList.remove(note);
            }
        } else if (note.getTransferUnit() != null) {
            if (notesEnc.remove(note.getTransferUnit().getRoot()) != note) {
                throw new IllegalArgumentException("Note not found in encapsulation note hashMap!");
            }
        } else {
            throw new IllegalArgumentException("Illegal instance of Note: both layer and transferUnit references were null. How did you even construct this monster? :)");
        }
    }
    
    /**
     * Returns the first note coming BEFORE the given time on the given layer.
     * This method is used in the UI to find the note to be shown at
     * a particular time.
     *
     * @param layer
     * @param time
     * @param inclusive whether the timestamp is inclusive or not
     * @return note or null if there is no note <b>before</b> the specified time
     */
    public Note getNoteTimeLayerPrev(Layer layer, float time, boolean inclusive) {
        //get an iterator pointing to the end of the layer specific list
        ListIterator iter = getNotesForLayer(layer).listIterator(getNotesForLayer(layer).size());
        while (iter.hasPrevious()) {
            //We are iterating the list from back to front
            Note noteElement = (Note) iter.previous();
            if (inclusive ? noteElement.getTime() <= time : noteElement.getTime() < time) {
                return noteElement;
            }
        }
        return null;
    }
    
    /**
     * Returns the first note coming AFTER the given time on the given layer.
     * This method is used to get the next "steppable" event for the global
     * animation event stepping architecture. This is needed for instance
     * to implement the note column, and stepping to notes.
     *
     * @param layer
     * @param time
     * @param inclusive whether the timestamp is inclusive or not
     * @return note or null if there is no note <b>after</b> the specified time
     */
    public Note getNoteTimeLayerNext(Layer layer, float time, boolean inclusive) {
        //get an iterator pointing to the start of the layer specific list
        ListIterator iter = getNotesForLayer(layer).listIterator();
        while (iter.hasNext()) {
            //We are iterating the list from front to back
            Note noteElement = (Note) iter.next();
            if (inclusive ? noteElement.getTime() >= time : noteElement.getTime() > time) {
                return noteElement;
            }
        }
        return null;
    }
    
    
    /**
     * Returns a Linked list (Collection) of all the notes corresponding to the requested layer.
     *
     * @param layer
     */
    public LinkedList getNotesForLayer(Layer layer) {
        LinkedList notes = (LinkedList)notesMsc.get(layer);
        if (notes == null) {
            notes = new LinkedList();
            notesMsc.put(layer,notes);
        }
        return notes;
    }
    
    /**
     * Returns the notes corresponding to the requested layer where the times
     * are between startTime and endTime.
     *
     * @param layer
     * @param startTime
     * @param endTime
     * @return Collection A LinkedList of notes in the given time range.
     */
    public Collection getNotesInTimeRange(Layer layer, float startTime, float endTime) {
        LinkedList relevantLayerList = getNotesForLayer(layer);
        Collection notes = new LinkedList();
        for (ListIterator iter = relevantLayerList.listIterator(); iter.hasNext();) {
            Note note = (Note) iter.next();
            if (note.getTime() >= startTime && note.getTime() <= endTime) {
                notes.add(note);
                //for "extra performance", we jump out if we have gone past the endtime
                //I hate breaks. Jari made me. -jonathan :)
            }else if (note.getTime() > endTime) {
                break;
            }
        }
        return notes;
    }
    
    /**
     * Adds an encapsulation note that is attached to a specific TransferUnit.
     *
     * @param noteText
     */
    public void addNoteEnc(Note note) {
        notesEnc.put(note.getTransferUnit().getRoot(), note);
    }
    
    /**
     * Returns the encapsulation note attached to the given TransferUnit.
     *
     * @param transferUnit
     * @return Note if exists, null otherwise
     */
    public Note getNoteEnc(TransferUnit transferUnit) {
        return (Note) notesEnc.get(transferUnit.getRoot());
    }
    
    /**
     * This method is used to save all notes to
     * a scenario file.
     */
    public Object getData() {
        return this;
    }
    
    /**
     * Used to load notes from a file. Assigns the 2 maps
     * from the loaded object to this object.
     */
    public void setData(Object o) {
        this.notesMsc = ((NoteManager)o).getNotesMsc();
        this.notesEnc = ((NoteManager)o).getNotesEnc();
    }
    
    /**
     * Returns the map where MSC notes are stored, in linked lists, one per layer.
     * This method is basically used only for retrieving a NoteManager instance from a
     * scenario file.
     * @return Map of MSC notes, with one LinkedList per layer.
     */
    public Map getNotesMsc() {
        return this.notesMsc;
    }
    
    /**
     * Returns the map where ENC notes are stored, in linked lists, one per layer.
     * This method is basically used only for retrieving a NoteManager instance from a
     * scenario file.
     * @return Map of ENC notes, keys are units.
     */
    public Map getNotesEnc() {
        return this.notesEnc;
    }
    
}