package fi.helsinki.dacopan.model;

/**
 * A {@link StepIterator} for stepping thru a whole scenario,
 * including both protocol events and notes.
 * 
 * @version $Id: ScenarioStepIterator.java,v 1.6 2004/05/27 07:27:10 jaarnial Exp $
 */
public class ScenarioStepIterator implements StepIterator {

    /** current timestamp */
    private float current;
    
    /** layer we're stepping thru */
    private Layer layer;
    
    /** end time of the layer */
    private float endTime;
    
    /** iterator for protocol events */
    private StepIterator events;
    
    /** are we in sync with the events iterator? */
    private boolean inSyncWithEvents = true;
    
    /** note manager instance */
    private NoteManager noteManager;

    /**
     * Creates the iterator for stepping thru the specified layer with the given
     * events and notes.
     */
    public ScenarioStepIterator(Layer layer, DataView view, NoteManager noteManager) {
        this.layer = layer;
        this.events = view.getStepIterator(layer);
        this.endTime = view.getEndTimeForLayer(layer);
        this.noteManager = noteManager;
        current = events.current();
    }

    public boolean hasNext() {
        return current < endTime;
    }

    public boolean hasPrevious() {
        return current > 0.0f;
    }

    public float next() {
        return moveToNext(
                current,
                inSyncWithEvents ? events.next() : events.getNextForTime(current),
                noteManager.getNoteTimeLayerNext(layer, current, false)
        );
    }

    public float previous() {
        return moveToPrevious(
                current,
                inSyncWithEvents ? events.previous() : events.getPreviousForTime(current),
                noteManager.getNoteTimeLayerPrev(layer, current, false)
        );
    }

    public float first() {
        return moveTo(events.first(), true);
    }

    public float last() {
        return moveTo(events.last(), true);
    }

    public float current() {
        return current;
    }

    public float getNextForTime(float time) {
        return moveToNext(
                time,
                events.getNextForTime(time),
                noteManager.getNoteTimeLayerNext(layer, time, false)
        );
    }

    public float getPreviousForTime(float time) {
        return moveToPrevious(
                time,
                events.getPreviousForTime(time),
                noteManager.getNoteTimeLayerPrev(layer, time, false)
        );
    }
    
    private float moveToNext(float now, float nextEvent, Note nextNote) {
        if(nextNote == null || nextEvent <= nextNote.getTime()) {
            // next protocol event is closest to current, use it
            moveTo(nextEvent, true);
        } else {
            // next note is closest to current, use it
            moveTo(nextNote.getTime(), false);
        }
        return current;
    }
    
    private float moveToPrevious(float now, float prevEvent, Note prevNote) {
        if(prevNote == null || prevEvent >= prevNote.getTime()) {
            // previous protocol event is closest to current, use it
            moveTo(prevEvent, true);
        } else {
            // prev note is closest to current, use it
            moveTo(prevNote.getTime(), false);
        }
        return current;
    }
    
    private float moveTo(float time, boolean isEvent) {
        current = time;
        inSyncWithEvents = isEvent;
        return current;
    }

}
