package fi.helsinki.dacopan.model;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * A unit of transfer (packet, segment etc) for a specific protocol.
 * 
 * @version $Id: TransferUnit.java,v 1.20 2004/05/24 20:25:23 jaarnial Exp $
 */
public class TransferUnit implements Comparable, Identifiable {
    
    /** unique identifier */
    protected long id;
    
    /** parent unit, which this unit is a part of (can be null) */
    protected TransferUnit parent;
    
    /** units that this unit consists of (can be null) */
    protected List children;
    
    /** protocol fields and other values associated with this unit, mapped by name */
    protected Map values = new HashMap();
    
    /** the protocol this unit belongs to */
    protected Protocol protocol;
    
    /** the flow this units belongs to (might be null) */
    protected Flow flow;
    
    /** source host */
    protected Host source;
    
    /** destination host */
    protected Host destination;
    
    /** time when sending of unit started */
    protected float sendStart;
    
    /** time when unit was fully sent */
    protected float sendEnd;
    
    /** time when receiving this unit started */
    protected float receiveStart;
    
    /** time when unit was fully received */
    protected float receiveEnd;
    
    /** whether this unit was dropped or not */
    protected boolean dropped;
    
    /**
     * @see TransferUnit#TransferUnit(long, Map, Protocol, Flow, Host, Host, float, float, boolean)
     */
    protected TransferUnit(long id, Map values, Protocol protocol, Flow flow, Host source, Host destination, boolean dropped) {
        if(values == null || protocol == null || source == null || destination == null) {
            throw new IllegalArgumentException("Values, protocol, source and destination must be non-null");
        }
        this.id = id;
        this.protocol = protocol;
        this.flow = flow;
        this.source = source;
        this.destination = destination;
        this.dropped = dropped;
        addInitialValues(values);
    }
    
    /**
     * Constructor for lower-level units with their own timestamps.
     * 
     * @param id unique id of the unit
     * @param values map (String->String) of variable names and values of this unit (not specific to any host)
     * @param protocol of the unit
     * @param flow (protocol-specific connection) of the unit
     * @param source host
     * @param destination host
     * @param sendStart time when sending of unit started
     * @param receiveEnd time when unit was fully received
     * @param dropped whether the unit was dropped
     */
    public TransferUnit(long id, Map values, Protocol protocol, Flow flow, Host source, 
        Host destination, float sendStart, float receiveEnd, boolean dropped) 
    {
        this(id, values, protocol, flow, source, destination, dropped);
        if(receiveEnd < sendStart) {
            throw new IllegalArgumentException("Receive time can't be smaller than send time for unit " + id);
        }
        this.sendStart = sendStart;
        this.sendEnd = sendStart;
        this.receiveStart = receiveEnd;
        this.receiveEnd = receiveEnd;
        assignChildren();
    }

    /**
     * Constructor for higher level units with children.
     * 
     * @see TransferUnit#TransferUnit(long, Map, Protocol, Flow, Host, Host, float, float, boolean)
     * @param children list of children units
     */
    public TransferUnit(long id, Map values, Protocol protocol, Flow flow, Host source, Host destination, List children, boolean dropped) {
        this(id, values, protocol, flow, source, destination, dropped);
        if(children == null) {
            throw new IllegalArgumentException("Children can't be null when no timestamps specified");
        }
        this.children = children;
        assignChildren();
        calculateTimestamps();
    }
    
    public long getId() {
        return id;
    }
    
    /**
     * Adds a variable value to this unit.
     * 
     * @param name name of the variable
     * @param host the value's specific to, use <code>null</code> if none
     * @param value of the variable
     */
    public void addVariableValue(String name, Host host, String value) {
        Value valueObject = new Value(value, host == source ? value : null, host == destination ? value : null);
        values.put(name, valueObject);
    }

    /**
     * @see #addVariableValue(String, Host, String)
     */
    public void addVariableValue(VariableDefinition def, Host host, String value) {
        addVariableValue(def.getRawName(), host, value);
    }
    
    /**
     * Gets the value of the given variable (ignoring if it's host-specific).
     */
    public String getValue(VariableDefinition variable) {
        return getValue(variable, null);
    }
    
    /**
     * Gets the value of the given variable (specific to the source host).
     */
    public String getValueSend(VariableDefinition variable) {
        return getValue(variable, source);
    }

    /**
     * Gets the value of the given variable (specific to the destination host).
     */
    public String getValueReceive(VariableDefinition variable) {
        return getValue(variable, destination);
    }

    /**
     * Returns the value of the given variable.
     * 
     * @param variable whose value to return
     * @param host to whom the value is specific (if any)
     * @return the value, null if none
     */
    public String getValue(VariableDefinition variable, Host host) {
        if(variable.getProtocol().equals(this.protocol)) {
            Value v = (Value)values.get(variable.getRawName());
            if(v != null) {
                if(host == null) {
                    return v.value;
                } else if(host == source) {
                    return v.sourceValue;
                } else if(host == destination) {
                    return v.destinationValue;
                }
            }
        }
        return null;
    }

    /**
     * Returns the protocol of this unit.
     */
    public Protocol getProtocol() {
        return protocol;
    }

    /**
     * Returns the flow this unit belongs to (may be null).
     */
    public Flow getFlow() {
        return flow;
    }

    /**
     * Returns the source host of this unit.
     */
    public Host getSource() {
        return source;
    }
    
    /**
     * Returns the destination host of this unit.
     */
    public Host getDestination() {
        return destination;
    }

    /**
     * Returns the children units of this unit (ordered by {@link #compareTo(Object) default ordering}).
     */
    public List getChildren() {
        return children;
    }

    /**
     * Returns the time this unit was sent.
     */
    public float getSendStart() {
        return sendStart;
    }

    /**
     * Returns the time this unit was fully received.
     */
    public float getReceiveEnd() {
        return receiveEnd;
    }

    /**
     * Returns the time the sending of this unit was completed.
     * If the unit has no children, this is always the same as
     * the value returned from {@link #getSendStart()}.
     */
    public float getSendEnd() {
        return sendEnd;
    }

    /**
     * Returns the time the destination host started to receive this unit.
     */
    public float getReceiveStart() {
        return receiveStart;
    }

    /**
     * Returns the parent of this unit (may be null)
     */
    public TransferUnit getParent() {
        return parent;
    }

    /**
     * Returns the root of the TransferUnit tree this unit belongs to.
     * 
     * @return the root unit, <code>this</code> if parent is null
     */
    public TransferUnit getRoot() {
        TransferUnit root = this;
        while(root.getParent() != null) {
            root = root.getParent();
        }
        return root;
    }

    /**
     * Sets the parent unit of this unit. The only setter method
     * in this otherwise immutable class. Needed because information
     * on the parent unit is not available when a unit is created
     * in a bottom-up approach.
     */
    public void setParent( TransferUnit parent ) {
        this.parent = parent;
    }

    /**
     * Returns true if this unit was dropped.
     */
    public boolean isDropped() {
        return dropped;
    }
    
    /**
     * Returns true if unit's sending and receiving are both just points in time and not time
     * intervals where start and end times are not equal.
     */
    public boolean isOneUnit() {
        return ((sendStart == sendEnd) && (receiveStart == receiveEnd));
    }
    
    /**
     * TransferUnits are ordered primarily by sendStart, then by receiveEnd.
     */
    public int compareTo(Object o) {
        TransferUnit other = (TransferUnit)o;
        int diff = Float.compare(sendStart, other.sendStart);
        if(diff == 0) {
            diff = Float.compare(receiveEnd, other.receiveEnd);
        }
        return diff;
    }

    public boolean equals(Object o) {
        if(this == o) return true;
        if(!(o instanceof TransferUnit)) return false;
        TransferUnit other = (TransferUnit)o;
        // compare all other fields apart from parent/children:
        // they might cause cirular references and general nastiness
        if(other.id != id) return false;
        if(!other.values.equals(values)) return false;
        if(!other.protocol.equals(protocol)) return false;
        if(flow != null ? !flow.equals(other.flow) : other.flow != null) return false;
        if(!other.source.equals(source)) return false;
        if(!other.destination.equals(destination)) return false;
        if(other.sendStart != sendStart) return false;
        if(other.sendEnd != sendEnd) return false;
        if(other.receiveStart != receiveStart) return false;
        if(other.receiveEnd != receiveEnd) return false;
        return true;
    }

    public String toString() {
        return "TransferUnit { U" + id + "/" + sendStart + "-" + receiveEnd + "/P" + protocol.getId() +
            "/H" + source.getId() + "->H" + destination.getId() +
            (flow != null ? "/F" + flow.getId() : "") + " " + values + " }";
    }
    
    // private/protected methods follow
    
    /**
     * Calculates the timestamps of this unit by looking at its children.
     */
    protected void calculateTimestamps() {
        sendStart = -1;
        receiveStart = -1;
        for(Iterator i = children.iterator(); i.hasNext(); ) {
            TransferUnit unit = (TransferUnit)i.next();
            if(sendStart == - 1 || unit.getSendStart() < sendStart) {
                sendStart = unit.getSendStart();
            }
            if(unit.getSendEnd() > sendEnd) {
                sendEnd = unit.getSendEnd();
            }
            if(receiveStart == -1 || unit.getReceiveStart() < receiveStart) {
                receiveStart = unit.getReceiveStart();
            }
            if(unit.getReceiveEnd() > receiveEnd) {
                receiveEnd = unit.getReceiveEnd();
            }
        }
    }

    /**
     * Assigns itself as the parent of its children.
     */
    protected void assignChildren() {
        if(children == null) {
            this.children = Collections.EMPTY_LIST;
            return;
        }
        for(Iterator i = children.iterator(); i.hasNext(); ) {
            TransferUnit child = (TransferUnit)i.next();
            child.setParent(this);
        }
        Collections.sort(children);
    }
    
    /**
     * Called from the constructor, adds all the values in the
     * map given by the user to the actual value map (makes the values
     * not specific to any host).
     */
    protected void addInitialValues(Map values) {
        for(Iterator i = values.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry)i.next();
            this.values.put((String)entry.getKey(), new Value((String)entry.getValue(), null, null));
        }
    }
    

    /**
     * Wraps a variable value and the info on any hosts it's specific to.
     */
    private final static class Value {
        private String value;
        private String sourceValue;
        private String destinationValue;
        private Value(String v, String s, String d) {
            value = v;
            sourceValue = s;
            destinationValue = d;
        }
        public boolean equals(Object o) {
            if(!(o instanceof Value)) return false;
            Value v = (Value)o;
            if(value != null ? !value.equals(v.value) : v.value != null) return false;
            return true;
        }
    }

}
