package fi.helsinki.dacopan.model;

import java.util.HashMap;
import java.util.Map;

import fi.helsinki.dacopan.Localization;

/**
 * Definition of a dynamic variable associated with a {@link Protocol protocol},
 * whose value can change from one {@link TransferUnit unit} to another. Each
 * variable is categorized to a specific {@link Scope scope}.
 * 
 * @see fi.helsinki.dacopan.model.StaticVariable
 * @version $Id: VariableDefinition.java,v 1.12 2004/05/11 09:21:05 aarniala-dacopan Exp $
 */
public class VariableDefinition {
    
    /** scope of a {@link TransferUnit unit} -specific variable (e.g. transition time) */
    public static final Scope UNIT = new Scope("unit");
    
    /** scope of a {@link Protocol protocol} field (e.g. TCP sequence number) */
    public static final Scope UNIT_FIELD = new Scope("unit-field");
    
    /** scope of a {@link Flow flow} -specific variable (e.g. TCP state) */
    public static final Scope FLOW = new Scope("flow");

    /** name of the variable */
    protected String name;
    
    /** scope of the variable */
    protected Scope scope;
    
    /** the protocol that this variables is associated with */
    protected Protocol protocol;

    /**
     * Creates a new variable definition with the given name, scope and protocol.
     */
    public VariableDefinition(String name, Scope scope, Protocol protocol) {
        if(name == null || scope == null || protocol == null) {
            throw new IllegalArgumentException("Name, scope and protocol required");
        }
        this.name = name;
        this.scope = scope;
        this.protocol = protocol;
    }
    
    /**
     * Returns the non-localized name of the variable.
     */
    public String getRawName() {
        return name;
    }
    
    public String getName() {
        return Localization.getString("model.variable."+protocol.name+"."+name+".abbr");
    }

    public String getFullName() {
        return Localization.getString("model.variable."+protocol.name+"."+name+".full");
    }
   
    public Scope getScope() {
        return scope;
    }
    
    public Protocol getProtocol() {
        return protocol;
    }

    public boolean isUnitVariable() {
        return scope.equals(UNIT);
    }

    public boolean isFlowVariable() {
        return scope.equals(FLOW);
    }

    public boolean isProtocolField() {
        return scope.equals(UNIT_FIELD);
    }
        
    public boolean equals(Object o) {
        if(this == o) return true;
        if(!(o instanceof VariableDefinition)) return false;
        VariableDefinition other = (VariableDefinition)o;
        if(!other.name.equals(name)) return false;
        if(!other.scope.equals(scope)) return false;
        if(other.protocol.id != protocol.id) return false;
        // note: see StaticVariable.equals() for an explanation
        // why we don't want to compare the protocol here
        return true;
    }
    
    public int hashCode() {
        int result;
        result = (int) (protocol.id ^ (protocol.id >>> 32));
        result = name.hashCode();
        result = 29 * result + scope.hashCode();
        return result;
    }
    
    public String toString() {
        return "{" + name + "/" + scope + "}";
    }
    
    /**
     * The scope of a variable.
     */
    public static final class Scope {

        /** registry of scopes */
        private static Map registry = new HashMap();

        /**
         * Returns the Scope object for the given name, if any.
         */
        public static final Scope forName(String name) {
            return (Scope)registry.get(name);
        }
        
        private String name;

        /**
         * Private constructor to ensure uninstantiability.
         */
        private Scope(String name) {
            this.name = name;
            registry.put(name, this);
        }
        
        /**
         * Returns the name of this scope.
         */
        public String getName() {
            return name;
        }
        
        public String toString() {
            return name;
        }
        
        public boolean equals(Object o) {
            return (o instanceof Scope && ((Scope)o).name.equals(name));
        }
        
        public int hashCode() {
            return name.hashCode();
        }
        
    }

}
