package fi.helsinki.dacopan.pef;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Flow;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.Link;
import fi.helsinki.dacopan.model.Protocol;
import fi.helsinki.dacopan.model.StaticVariable;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;
import fi.helsinki.dacopan.ui.ProgressIndicator;
import fi.helsinki.dacopan.ui.ProgressIndicator.NoopIndicator;

/**
 * The default, DOM-based implementation of the PEF reader.
 * 
 * @version $Id: XMLProtocolEventsReader.java,v 1.17 2004/05/24 20:25:23 jaarnial Exp $
 */
public class XMLProtocolEventsReader implements ProtocolEventsReader, EntityResolver {

    /** system id of the DTD that PEF instances should use */
    public static final String DTD_SYSTEM_ID = "http://www.cs.helsinki.fi/group/dacopan/events.dtd";
    
    /** the classpath resource where the DTD should reside */
    public static final String DTD_RESOURCE = "/events.dtd";
    
    /** source DOM document used to populate the view */
    protected Document document;
    
    /** the view to be populated */
    protected DataView view;

    public void read(Reader reader, DataView view) {
        read(reader, view, null);
    }
    
    /**
     * Reads in the contents of the reader and notifies the given indicator
     * of any progress being made.
     */
    public void read(Reader reader, DataView view, ProgressIndicator indicator) {
        if(indicator == null) {
            indicator = new NoopIndicator();
        }
        this.document = parseDocument(reader);
        indicator.setProgress(20.0f);
        this.view = view;
        processHosts();
        processFlows();
        processLinks();
        processLayersAndProtocols();
        processVariables();
        indicator.setProgress(30.0f);
        processUnits(indicator);
    }
    
    /**
     * Provides special resolving for the PEF DTD: it is loaded from the classpath.
     */
    public InputSource resolveEntity(String publicId, String systemId)
            throws SAXException, IOException {
        if (systemId.equals(DTD_SYSTEM_ID)) {
            InputStream is = getClass().getResourceAsStream(DTD_RESOURCE);
            if(is == null) {
                throw new RuntimeException(
                        "Classpath resource " + DTD_RESOURCE + " not found. \n" +
                        "If this exception occurs while running the test \n" +
                        "suite, make sure you have added the directory \n" +
                        "dacopan/documents/design to the classpath."
                );
            }
            return new InputSource(is);
        }
        return null;
    }

    /**
     * Parses the contents of the reader to a DOM document.
     */
    protected Document parseDocument(Reader reader) {
        try {
            // set up the factory, with validation
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            factory.setValidating(true);
            // then the builder, provide our own entity resolver
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.setEntityResolver(this);
            builder.setErrorHandler(new HaltOnErrorHandler());
            Document doc = builder.parse(new InputSource(reader));
            if(!DTD_SYSTEM_ID.equals(doc.getDoctype().getSystemId())) {
                throw new ProtocolEventsDataException("Wrong doctype");
            }
            return doc;
        } catch (Exception e) {
            if(e instanceof ProtocolEventsDataException) {
                throw (ProtocolEventsDataException)e;
            } else {
                throw new ProtocolEventsDataException("Reading PEF failed: " + e.getMessage(), e);
            }
        }
    }

    protected void processHosts() {
        NodeList hosts = document.getElementsByTagName("host");
        for(int i = 0; i < hosts.getLength(); i++) {
            Node host = hosts.item(i);
            view.addHost(new Host(
                    getId(host),
                    getStringAttribute(host, "ip"),
                    getStringAttribute(host, "hostname")
            ));
        }
        if(view.getHosts().size() < 2) {
            // this restriction is imposed by the MSC animation
            throw new IllegalArgumentException("A PEF must contain at least 2 hosts");
        }
    }
    
    protected void processFlows() {
        NodeList flows = document.getElementsByTagName("flow");
        for(int i = 0; i < flows.getLength(); i++) {
            Element flowElement = (Element)flows.item(i);
            Flow flow = new Flow(
                    getId(flowElement),
                    view.getHostById(getId(flowElement, "host1")),
                    getIntegerAttribute(flowElement, "port1"),
                    view.getHostById(getId(flowElement, "host2")),
                    getIntegerAttribute(flowElement, "port2")
            );
            view.addFlow(flow);
        }
    }
    
    protected void processLinks() {
        NodeList links = document.getElementsByTagName("link");
        for(int i = 0; i < links.getLength(); i++) {
            Node linkElement = links.item(i);
            Link link = new Link(
                    getId(linkElement),
                    view.getHostById(getId(linkElement, "host1")),
                    view.getHostById(getId(linkElement, "host2"))
            );
            view.addLink(link);
        }
    }
    
    protected void processLayersAndProtocols() {
        NodeList layers = document.getElementsByTagName("layer");
        int order = 0;
        for(int i = 0; i < layers.getLength(); i++) {
            Element layerEl = (Element)layers.item(i);
            Layer layer = new Layer(getId(layerEl), getStringAttribute(layerEl, "name"), order++);
            view.addLayer(layer);
            processProtocols(layer, layerEl.getElementsByTagName("protocol"));
        }
    }
    
    private void processProtocols(Layer parent, NodeList protocols) {
        for(int i = 0; i < protocols.getLength(); i++) {
            // find out layer by looking at parent
            Element protocolNode = (Element)protocols.item(i);
            Protocol protocol = new Protocol(
                    getId(protocolNode),
                    getStringAttribute(protocolNode, "name"),
                    parent
            );
            view.addProtocol(protocol);
            parent.getProtocols().add(protocol);
        }
    }
    
    protected void processVariables() {
        NodeList constants = document.getElementsByTagName("constant");
        for(int i = 0; i < constants.getLength(); i++) {
            Node constantNode = constants.item(i);
            Host host = view.getHostById(getId(constantNode, "host"));
            Link link = view.getLinkById(getId(constantNode, "link"));
            Protocol protocol = view.getProtocolById(getId(constantNode, "protocol"));
            if(host == null && link == null && protocol == null) {
                throw new ProtocolEventsDataException("Constants require owner(s)");
            }
            StaticVariable constant = new StaticVariable(getStringAttribute(constantNode, "name"), getText(constantNode), host, link, protocol);
            if(host != null) {
                host.getStaticVariables().add(constant);
            }
            if(link != null) {
                link.getStaticVariables().add(constant);
            }
            if(protocol != null) {
                protocol.getStaticVariables().add(constant);
            }
        }
        NodeList variables = document.getElementsByTagName("variable");
        for(int i = 0; i < variables.getLength(); i++) {
            Node variableNode = variables.item(i);
            long[] protocolIds = getIdRefs(variableNode, "protocol");
            for (int j = 0; j < protocolIds.length; j++) {
                Protocol proto = view.getProtocolById(protocolIds[j]);
                VariableDefinition def = new VariableDefinition(
                        getStringAttribute(variableNode, "name"),
                        VariableDefinition.Scope.forName(getStringAttribute(variableNode, "scope")),
                        proto
                );
                proto.getVariables().add(def);
            }
        }
    }
    
    protected void processUnits(ProgressIndicator indicator) {
        NodeList sentUnits = document.getElementsByTagName("unit_sent");
        int numUnits = sentUnits.getLength();
        if(numUnits > 0) {
            indicator.setStep(70.0f / numUnits);
        }
        for(int item = 0; item < numUnits; item++) {
            Element sent = (Element)sentUnits.item(item);
            long id = getId(sent);
            if(view.getUnitById(id) == null) {
                processStartElement(id, sent, indicator);
            }
        }
    }
    
    private TransferUnit processStartElement(long id, Element start, ProgressIndicator indicator) {
        Element end = findEndTag(id, start);
        Host source = view.getHostById(getId(start, "source"));
        Host destination = view.getHostById(getId(start, "destination"));
        Flow flow = null;
        if(getStringAttribute(start, "flow") != null) {
            flow = view.getFlowById(getId(start, "flow"));
        }
        Protocol protocol = view.getProtocolById(getId(start, "protocol"));
        boolean dropped = end.getNodeName().equals("unit_dropped");
        TransferUnit unit;
        long[] ids = getIdRefs(start, "children");
        if(ids.length > 0) {
            List children = new ArrayList();
            for(int ch = 0; ch < ids.length; ch++) {
                long childId = ids[ch];
                TransferUnit child = view.getUnitById(childId);
                if(child == null) {
                	// child hasn't been processed yet, do so now
                	Element childStart = document.getElementById("U" + childId);
                	child = processStartElement(childId, childStart, indicator);
                }
                children.add(child);
            }
            unit = new TransferUnit(id, Collections.EMPTY_MAP, protocol, flow, source, destination, children, dropped);
        } else {
            // no children, use timestamps as defined in the unit_sent node
            float sendStart = getFloatAttribute(start, "time");
            float receiveEnd = getFloatAttribute(end, "time");
            if(receiveEnd < sendStart) {
                throw new ProtocolEventsDataException("Unit " + id + "'s receive time (" + receiveEnd + ") is smaller than send time (" + sendStart + ")");
            }
            unit = new TransferUnit(id, Collections.EMPTY_MAP, protocol, flow, source, destination, sendStart, receiveEnd, dropped);
        }
        // add the variable values specific to the sending host
        addValues(unit, start, source);
        /* TODO later, if values can be assigned to receiving host
         * only, something like this will do the trick:
         * addValues(unit, end, destination);
         */
        view.addTransferUnit(unit, false);
        indicator.step();
        return unit;
    }
    
    private Element findEndTag(long id, Node start) {
        Node next = start;
        while((next = next.getNextSibling()) != null) {
            if(next.getNodeType() == Node.ELEMENT_NODE && getId(next) == id
                    && (next.getNodeName().equals("unit_received") || next.getNodeName().equals("unit_dropped")))
            {
                return (Element)next;
            }
        }
        throw new ProtocolEventsDataException("No corresponding unit_received or unit_dropped for unit_sent with id " + id);
    }
    
    private void addValues(TransferUnit unit, Element base, Host host) {
        NodeList valueElements = base.getElementsByTagName("value");
        for(int i = 0; i < valueElements.getLength(); i++) {
            Node valueElement = valueElements.item(i);
            unit.addVariableValue(
                    getStringAttribute(valueElement, "name"),
                    host,
                    getText(valueElement)
            );
        }
    }
    
    /**
     * Returns the id value of the given node.
     */
    protected long getId(Node node) {
        return getId(node, "id");
    }
    
    /**
     * Returns the id value of the given attribute of the given node.
     */
    protected long getId(Node node, String name) {
        String id = getStringAttribute(node, name);
        if(id == null || id.length() < 2) {
            return -1;
        }
        return Long.parseLong(id.substring(1));
    }

    /**
     * Returns the value of the attribute with the given name.
     */
    protected String getStringAttribute(Node node, String name) {
        Node v = node.getAttributes().getNamedItem(name);
        if(v != null) {
            return v.getNodeValue();
        }
        return null;
    }
    
    /**
     * Gets IDREFS in an array.
     */
    protected long[] getIdRefs(Node node, String name) {
        String value = getStringAttribute(node, name);
        if(value != null) {
            String[] values = value.split("\\s+");
            long[] idRefs = new long[values.length];
            for (int i = 0; i < values.length; i++) {
                idRefs[i] = Long.parseLong(values[i].substring(1));
            }
            return idRefs;
        } else {
            return new long[0];
        }
    }
    
    /**
     * Returns the attribute value as an integer.
     */
    protected int getIntegerAttribute(Node node, String name) {
        String s = getStringAttribute(node, name);
        if(s == null) {
            throw new ProtocolEventsDataException("Attribute '" + name + "' is null in " + node + ", can't parse to int");
        }
        return Integer.parseInt(s);
    }

    /**
     * Returns the attribute value as a float.
     */
    protected float getFloatAttribute(Node node, String name) {
        String s = getStringAttribute(node, name);
        if(s == null) {
            throw new ProtocolEventsDataException("Attribute '" + name + "' is null in " + node + ", can't parse to float");
        }
        return Float.parseFloat(s);
    }
    
    /**
     * Returns the text content of the given node.
     */
    protected String getText(Node node) {
        NodeList children = node.getChildNodes();
        for(int i = 0; i < children.getLength(); i++) {
            switch(children.item(i).getNodeType()) {
                case Node.TEXT_NODE:
                case Node.CDATA_SECTION_NODE:
                return children.item(i).getNodeValue();
            }
        }
        return null;
    }
    
    /**
     * A SAX ErrorHandler that halts on any errors.
     */
    private class HaltOnErrorHandler implements ErrorHandler {
        
        /**
         * Creates a more descriptive error message from the SAX error, throwing it if ignore is false.
         */
        private void reportError(SAXParseException e, boolean ignore) throws SAXException {
            String msg = "Error parsing PEF (line " + e.getLineNumber() + ", column "
            	+ e.getColumnNumber() + ") " + e.getMessage();
            System.err.println(msg);
            if(!ignore) {
                throw new SAXParseException(msg, null, e);
            }
        }
        
        public void error(SAXParseException e) throws SAXException {
            reportError(e, false);
        }
        
        public void fatalError(SAXParseException e) throws SAXException {
            reportError(e, false);
        }
        
        public void warning(SAXParseException e) throws SAXException {
            reportError(e, true);
        }
        
    }
}