/*
 * MSCPanel.java
 *
 * Created on 29. huhtikuuta 2004, 16:24
 */

package fi.helsinki.dacopan.ui;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.JPanel;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Note;
import fi.helsinki.dacopan.model.StaticVariable;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseWheelListener;

/**
 *
 *
 */

public class MSCPanel extends fi.helsinki.dacopan.ui.AbstractAnimationPanel {
    
    private static final int MSC_VERTICAL_MARGIN = 16;
    private static final int MSC_SCROLLBAR_MULTIPLIER = 10000;
    private static final int MSC_SCROLLBAR_UNIT_INCREMENT_RATIO = 10;
    
    private static final int WIDTH_COLUMN_NOTES = 15;
    private static final int WIDTH_COLUMN_PER_CHARACTER = 8;
    private static final int WIDTH_DRAWING_AREA_MINIMUM = 100;
    
    private static final int MAX_DRAWABLE_VALUE_LENGTH = 150;
    
    /** Disables using antialias for drawing 'fat' units (fat is defined as unit.isOneUnit()
     * returning false).*/
    private static final boolean DISABLE_ANTIALIAS_FOR_FAT_UNITS = true;
    
    /** Uses optimized drawing in play mode (advanceMode == true). The optimized drawing means
     * caching image data in a buffer and then drawing slices on screen as the now line progresses.
     * The buffer is only refreshed when the screen needs to 'nudge'.
     * 
     * <p>Note that currently the optimization scheme is rather simple. The buffer data is
     * refreshed completely, instead of utilizing the 2/3 of it that would be usable. Also,
     * the buffer might also be used to optimize drawing in scrolling mode. However, these
     * ideas are not currently implemented.
     *
     * <p>Note also that the optimized drawing changes slightly the way in which column values
     * and notes appear on screen. While stepping, the values for current event appear even if
     * they have to be drawn below the now line. This is not true in optimized play mode, because
     * the image is simply cropped at now line. However, as the now line is moving, noticing
     * the difference in play mode would normally be extremely difficult.
     */
    private static final boolean USE_OPTIMIZED_PLAYMODE_DRAWING = true;
    
    private static final Color COLOR_DROPPED_CROSS = Color.red;
    private static final int DROPPED_CROSS_WIDTH = 16;
    
    private static final Color COLOR_COLUMN_BORDERS = Color.gray;
    private static final Color COLOR_NOWLINE = Color.blue;
    private static final Color COLOR_COLUMN_VALUES = Color.black;
    private static final Color COLOR_EVENT_LINES = new Color(0x33, 0x33, 0x33);
    private static final Color COLOR_EVENT_AREA = new Color(204, 204, 75);
    private static final Color COLOR_EVENT_VALUES = Color.black;
    
    /** The margin left between the line representing a unit on MSC, and the corresponding
     * text presenting the header field values for that unit. This is in absolute pixels on
     * absolute, regardless of the angle of the text.
     */
    private static final int EVENT_VALUE_YDIFF = 3;
    
    private static final String NOTE_ICON_PATH = "/icon_note.png";

    private BufferedImage noteImage = null;
    {
        try {
            noteImage = ImageIO.read(getClass().getResourceAsStream(NOTE_ICON_PATH));
        } catch (IOException e) {
            System.out.println("MSC Panel trying to load note icon:");
            System.out.println(e);
            System.out.flush();
        }
    }
    
    private final SettingsMSC settings;
    private final DataView dataView;
    
    private final MSCPanel.MSCDrawingPanel drawingPanel;
    
    private final boolean drawBelowNowLine = false;
    
    private final boolean useAntiAlias;
    
    private boolean advanceMode = false;
    private boolean isScrolling = false;
    private boolean isSteppingBackwards = false;
    
    /* true -> false: when jScrollBar1.setValue is called
     * false -> true: jScrollBar1AdjustmentValueChanged method beginning, and
     * return from the method.
     *
     * effect: when the value is set programmatically, the first call to
     * jScrollBar1AdjustmentValueChanged will not be interpreted as user action
     */
    private boolean scrollBarValueSet = false;
    
    private float timeNow;
    
    private final float verticalMarginTime;
    
    
    private MSCColumnModel columnModel;
    
    /** Creates new form MSCPanel */
    public MSCPanel(
        MainFrame mainFrame, 
        SettingsMSC settings, 
        DataView dataView,
        float nowTime) 
    {
        super(mainFrame);
        
        if (settings == null) throw new IllegalArgumentException("Settings can't be null!");
        this.settings = settings;
        
        if (dataView == null) throw new IllegalArgumentException("DataView can't be null!");
        this.dataView = dataView;

        if (nowTime < 0f) throw new IllegalArgumentException("nowTime can't be negative!");
        this.timeNow = nowTime;
        
        useAntiAlias = mainFrame.getSettings().getAntiAliasing();
        
        drawingPanel = new MSCDrawingPanel();
        
        columnModel = new MSCColumnModel(getDrawableWidth(), settings.getVariablesColumns());

        verticalMarginTime = 1.0f / (settings.getVisualScale() / (float)MSC_VERTICAL_MARGIN);
        
        initComponents();

        jPanel2.setBorder(BorderFactory.createLineBorder(Color.gray));
        
        jScrollBar1.setMinimum(0);
        jScrollBar1.setMaximum((int)((dataView.getEndTime() + (2*verticalMarginTime))
            * MSC_SCROLLBAR_MULTIPLIER));
        
        drawingPanel.addMouseWheelListener(drawingPanel);
        jScrollBar1.addMouseWheelListener(drawingPanel);
        
        // set to true to prevent scrollbar from yanking to top
        scrollBarValueSet = true;
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        java.awt.GridBagConstraints gridBagConstraints;

        jPanel1 = new javax.swing.JPanel();
        jPanelTitle = new TitlePanel(Localization.getString("panel.msc.title"), settings.getLayer().getName());
        jPanelHosts = new javax.swing.JPanel();
        jLabelHostLeft = new javax.swing.JLabel();
        jLabelHostRight = new javax.swing.JLabel();
        jTableHostVariablesLeft = new javax.swing.JTable();
        jTableHostVariablesRight1 = new javax.swing.JTable();
        jPanel2 = new javax.swing.JPanel();
        jPanelColumnHeadingPanel = new ColumnHeadingPanel();
        jPanelDrawingPanel = drawingPanel;
        jScrollBar1 = new javax.swing.JScrollBar();

        setLayout(new java.awt.BorderLayout());

        jPanel1.setLayout(new java.awt.GridBagLayout());

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        jPanel1.add(jPanelTitle, gridBagConstraints);

        jPanelHosts.setLayout(new java.awt.GridBagLayout());

        jLabelHostLeft.setText(Localization.getString("panel.msc.host_a")+settings.getLeftHost().getIp());
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.insets = new java.awt.Insets(2, 4, 2, 4);
        jPanelHosts.add(jLabelHostLeft, gridBagConstraints);

        jLabelHostRight.setText(Localization.getString("panel.msc.host_b")+settings.getRightHost().getIp());
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.insets = new java.awt.Insets(2, 4, 2, 4);
        jPanelHosts.add(jLabelHostRight, gridBagConstraints);

        jTableHostVariablesLeft.setBorder(new javax.swing.border.LineBorder(new java.awt.Color(0, 0, 0)));
        jTableHostVariablesLeft.setModel(new HostVariableTableModel(settings.getLeftHost()));
        jTableHostVariablesLeft.setAutoResizeMode(javax.swing.JTable.AUTO_RESIZE_OFF);
        jTableHostVariablesLeft.setFocusable(false);
        jTableHostVariablesLeft.setMaximumSize(null);
        jTableHostVariablesLeft.setMinimumSize(new java.awt.Dimension(20, 20));
        jTableHostVariablesLeft.setPreferredSize(new java.awt.Dimension(60, 20));
        jTableHostVariablesLeft.setRowSelectionAllowed(false);
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.weightx = 0.5;
        gridBagConstraints.insets = new java.awt.Insets(2, 4, 2, 4);
        jPanelHosts.add(jTableHostVariablesLeft, gridBagConstraints);

        jTableHostVariablesRight1.setBorder(new javax.swing.border.LineBorder(new java.awt.Color(0, 0, 0)));
        jTableHostVariablesRight1.setModel(new HostVariableTableModel(settings.getRightHost()));
        jTableHostVariablesRight1.setAutoResizeMode(javax.swing.JTable.AUTO_RESIZE_OFF);
        jTableHostVariablesRight1.setFocusable(false);
        jTableHostVariablesRight1.setMaximumSize(null);
        jTableHostVariablesRight1.setMinimumSize(new java.awt.Dimension(20, 20));
        jTableHostVariablesRight1.setPreferredSize(new java.awt.Dimension(60, 20));
        jTableHostVariablesRight1.setRowSelectionAllowed(false);
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.weightx = 0.5;
        gridBagConstraints.insets = new java.awt.Insets(2, 4, 2, 4);
        jPanelHosts.add(jTableHostVariablesRight1, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.insets = new java.awt.Insets(2, 2, 2, 2);
        jPanel1.add(jPanelHosts, gridBagConstraints);

        jPanel2.setLayout(new java.awt.GridBagLayout());

        jPanelColumnHeadingPanel.setMinimumSize(new java.awt.Dimension(100, 20));
        jPanelColumnHeadingPanel.setPreferredSize(new java.awt.Dimension(100, 20));
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.SOUTH;
        gridBagConstraints.weightx = 1.0;
        jPanel2.add(jPanelColumnHeadingPanel, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        jPanel2.add(jPanelDrawingPanel, gridBagConstraints);

        jScrollBar1.addAdjustmentListener(new java.awt.event.AdjustmentListener() {
            public void adjustmentValueChanged(java.awt.event.AdjustmentEvent evt) {
                jScrollBar1AdjustmentValueChanged(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.VERTICAL;
        gridBagConstraints.weighty = 1.0;
        jPanel2.add(jScrollBar1, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        gridBagConstraints.insets = new java.awt.Insets(4, 4, 4, 4);
        jPanel1.add(jPanel2, gridBagConstraints);

        add(jPanel1, java.awt.BorderLayout.CENTER);

    }//GEN-END:initComponents

    private void jScrollBar1AdjustmentValueChanged(java.awt.event.AdjustmentEvent evt) {//GEN-FIRST:event_jScrollBar1AdjustmentValueChanged
        if (scrollBarValueSet) {
            scrollBarValueSet = false;
            return;
        }
        
        if (advanceMode) {
            drawingPanel.setScrollBarValue();
        } else {
            isScrolling = true;
            drawingPanel.repaint();
        }
    }//GEN-LAST:event_jScrollBar1AdjustmentValueChanged
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel jLabelHostLeft;
    private javax.swing.JLabel jLabelHostRight;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel jPanel2;
    private javax.swing.JPanel jPanelColumnHeadingPanel;
    private javax.swing.JPanel jPanelDrawingPanel;
    private javax.swing.JPanel jPanelHosts;
    private javax.swing.JPanel jPanelTitle;
    private javax.swing.JScrollBar jScrollBar1;
    private javax.swing.JTable jTableHostVariablesLeft;
    private javax.swing.JTable jTableHostVariablesRight1;
    // End of variables declaration//GEN-END:variables
    
    
    private final int getDrawableWidth() {
        return drawingPanel.getWidth() - drawingPanel.getInsets().left
            - drawingPanel.getInsets().right;
    }
    
    public void stepTo(float nowTime) {
        advanceMode = false;
        isScrolling = false;
        if (this.timeNow > nowTime) isSteppingBackwards = true;
        else isSteppingBackwards = false;
        this.timeNow = nowTime;
        drawingPanel.repaint();
    }
    
    public void advance(float step, float nowTime) {
        isScrolling = false;
        isSteppingBackwards = false;
        
        if ((!advanceMode) || (!USE_OPTIMIZED_PLAYMODE_DRAWING) ||
            drawingPanel.timeDrawingBottomChanges(nowTime) ||
            (!drawingPanel.isBufferedImageSizeValid())) 
        {
            advanceMode = true;
            this.timeNow = nowTime;
            drawingPanel.repaint();
        } else {
            drawingPanel.advanceDrawingTo(this.timeNow, nowTime);
            this.timeNow = nowTime;
        }
        
    }
    
    public void toPlayMode() {
        advanceMode = true;
        isScrolling = false;
        isSteppingBackwards = false;
    }
    
    public void toPauseMode() {
        if (advanceMode) {
            advanceMode = false;
            drawingPanel.setScrollBarValue();
        }
    }
    
    public MSCColumnModel getColumnModel() {
        if (getDrawableWidth() != columnModel.getTotalWidth()) {
            columnModel = new MSCColumnModel(getDrawableWidth(), settings.getVariablesColumns());
            return columnModel;
        } else {
            return columnModel;
        }
    }
    
private class MSCDrawingPanel extends JPanel implements MouseWheelListener {

    private float timeDrawingTop;
    private float timeDrawingBottom;

    /** The previously drawn time interval measured as the interval length on time axis from
     * the top of the drawing area to the bottom of the drawing area, regardless of the location
     * of the now-line.
     */
    private float prevDrawInterval;
    private float currentDrawInterval;
    
    private int previousDrawableWidth;
    private int previousDrawableHeight;

    private BufferedImage bufferedImage = null;
    
    private boolean isMouseDragging = false;
    
    /* Used to prevent handling events in a wrong order. */
    private boolean isMouseDraggingFinished = false;
    
    /* Used to prevent stupid behaviour if user touches >1 mouse buttons during drag.
     * The drag ends exactly when the button that was pressed first is released. */
    private int mouseDraggingButton = MouseEvent.NOBUTTON;

    private int paintCount = 0; // used only for debugging output

    public MSCDrawingPanel() {
        this.timeDrawingTop = timeNow;
        this.timeDrawingBottom = timeNow + ((float)getDrawableHeight() / settings.getVisualScale());
        this.prevDrawInterval = (float)getDrawableHeight() / settings.getVisualScale();
        setOpaque(true);
        this.setBackground(Color.WHITE);
        this.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent evt) {
                if (!advanceMode && (mouseDraggingButton == MouseEvent.NOBUTTON)) {
                    mouseDraggingButton = evt.getButton();
                    isMouseDragging = true;
                    isMouseDraggingFinished = false;
                    doMouseSelecting(evt.getY());
                }
            }
            
            public void mouseReleased(MouseEvent evt) {
                if (evt.getButton() == mouseDraggingButton) {
                    isMouseDraggingFinished = true;
                    mouseDraggingButton = MouseEvent.NOBUTTON;
                }
            }
        });
        this.addMouseMotionListener(new MouseMotionAdapter() {
            public void mouseDragged(MouseEvent evt) {
                if (!advanceMode && isMouseDragging && !isMouseDraggingFinished) {
                    doMouseSelecting(evt.getY());
                }
            }
        });
    }

    protected final int getNewScrollBarValue() {
        int newValue;
        if (timeDrawingTop < 0f) {
            newValue = 0;
        } else {
            newValue = (int)((timeDrawingTop + verticalMarginTime) 
                * MSC_SCROLLBAR_MULTIPLIER);
        }

        return newValue;
    }
    
    protected final void setScrollBarValue() {
        if (isScrolling) return;

        int newValue = getNewScrollBarValue();
        
        scrollBarValueSet = true;
        if (newValue != jScrollBar1.getValue()) {
            jScrollBar1.setValue(newValue);
        }
        scrollBarValueSet = false;
    }

    private final int getDrawableHeight() {
        return getHeight() - getInsets().top - getInsets().bottom;
    }
    
    /** Returns the new time representing the bottom of drawing area. Method is parameterized
     * to take the conceptual 'now time' as the parameter newTime, but otherwise the method relies
     * in some instance variables, such as advanceMode, isScrolling, currentDrawInterval and
     * prevDrawInterval. Method doesn't have any side-effects.
     */
    private float getTimeDrawingBottom(float newTime) {
        float newTimeDrawingBottom = timeDrawingBottom;
        
        if (isMouseDragging && !isMouseDraggingFinished) {
            // do nothing, we don't want to move display while dragging
        } else if (prevDrawInterval != currentDrawInterval) {
            // resize has happened
            if (newTime > currentDrawInterval-verticalMarginTime) {
                newTimeDrawingBottom = newTime + verticalMarginTime;
            } else {
                newTimeDrawingBottom = currentDrawInterval-verticalMarginTime;
            }
        } else if (advanceMode) {
            // we are in advance (play) mode
            if (newTime > timeDrawingBottom) {
                // is step the size of one 'nudge'?
                if (newTime < timeDrawingBottom + (currentDrawInterval / 3.0f)) {
                    // TODO implement sliding
                    newTimeDrawingBottom += currentDrawInterval / 3.0f;
                } else {
                    newTimeDrawingBottom = newTime + (currentDrawInterval / 3.0f);
                }
            } else if (newTime < timeDrawingTop) {
                // trying to play while now line is way above display, yank display to it
                newTimeDrawingBottom = newTime + (currentDrawInterval / 3.0f);
            } else {
                // do nothing
            }
        } else if (isScrolling) {
            newTimeDrawingBottom = ((float)jScrollBar1.getValue() / (float)MSC_SCROLLBAR_MULTIPLIER)
                + currentDrawInterval - verticalMarginTime;
        } else { // no advance mode, no resize, just stepping
            if (newTime < currentDrawInterval-verticalMarginTime) {
                newTimeDrawingBottom = currentDrawInterval-verticalMarginTime;
            } else {
                // check if we are still is visible lowest third
                if (newTime <= timeDrawingBottom-verticalMarginTime &&
                    newTime >= (timeDrawingTop + 
                    ((timeDrawingBottom-timeDrawingTop) * 2.0f / 3.0f)))
                {
                    // do nothing
                } else if (newTime >= timeDrawingBottom-verticalMarginTime) {
                    // stepped below bottom
                    newTimeDrawingBottom = newTime + 
                        (currentDrawInterval / 3.0f);
                } else if (isSteppingBackwards && newTime <= (timeDrawingTop + 
                    ((timeDrawingBottom-timeDrawingTop) * 2.0f / 3.0f)))
                {
                    // stepped above lowest one third
                    newTimeDrawingBottom = newTime + verticalMarginTime;
                } else {
                    // do nothing
                }
            }
        }
        
        float endTime = dataView.getEndTime();
        if (endTime > currentDrawInterval - (2f*verticalMarginTime) &&
            newTimeDrawingBottom > endTime + verticalMarginTime) 
        {
            newTimeDrawingBottom = endTime + verticalMarginTime;
        }
        
        return newTimeDrawingBottom;
    }
    
    protected final boolean timeDrawingBottomChanges(float newTime) {
        return timeDrawingBottom != getTimeDrawingBottom(newTime);
    }
    
    protected final boolean isBufferedImageSizeValid() {
        if ((bufferedImage == null) || 
            (bufferedImage.getWidth() != getDrawableWidth()) ||
            (bufferedImage.getHeight() != getDrawableHeight()))
        {
            return false;
        } else {
            return true;
        }
    }

    protected void updateDrawingTopBottomAndInterval(float nowTime) {
        currentDrawInterval = (float)getDrawableHeight() / settings.getVisualScale();
        timeDrawingBottom = getTimeDrawingBottom(nowTime);
        timeDrawingTop = timeDrawingBottom - currentDrawInterval;
    }

    /** Gets new scrollbar extent based on currentDrawInterval. */
    protected int getNewScrollBarExtent() {
        float usableDrawInterval = currentDrawInterval - (2f * verticalMarginTime);
        int newExtent;
        if (usableDrawInterval / dataView.getEndTime() > 1f) {
            newExtent = jScrollBar1.getMaximum();
        } else {
            newExtent = (int)((currentDrawInterval / dataView.getEndTime()) 
                * dataView.getEndTime() * MSC_SCROLLBAR_MULTIPLIER);
        }
        
        return newExtent;
    }
    
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);

        if (getDrawableWidth() == 0 || getDrawableHeight() == 0) return;

        Graphics2D g2Component = (Graphics2D)g;
        Graphics2D g2Buffer = null;
        
        if (advanceMode && USE_OPTIMIZED_PLAYMODE_DRAWING) {
            if (bufferedImage == null || !isBufferedImageSizeValid())
            {
                bufferedImage = g2Component.getDeviceConfiguration()
                    .createCompatibleImage(getDrawableWidth(), getDrawableHeight());
            }
        
            Graphics bufG = bufferedImage.getGraphics();
            bufG.setColor(this.getBackground());
            bufG.fillRect(0, 0, bufferedImage.getWidth(), bufferedImage.getHeight());
            
            g2Buffer = (Graphics2D)bufferedImage.getGraphics();
        }

        final MSCColumnModel columnModel = getColumnModel();

        updateDrawingTopBottomAndInterval(timeNow);

        // test if area has changed
        if (prevDrawInterval != currentDrawInterval) {
            // come out of scrolling
            isScrolling = false;
            isSteppingBackwards = false;

            // change scrollbar extent property
            int newExtent = getNewScrollBarExtent();
            
            if (jScrollBar1.getValue() + newExtent > jScrollBar1.getMaximum()) {
                jScrollBar1.setValue(jScrollBar1.getMaximum() - newExtent);
            }
            jScrollBar1.setVisibleAmount(newExtent);
            
            jScrollBar1.setUnitIncrement(jScrollBar1.getVisibleAmount() / 
                MSC_SCROLLBAR_UNIT_INCREMENT_RATIO);
            jScrollBar1.setBlockIncrement(jScrollBar1.getVisibleAmount());
        }
        
        float bottomCropTime;
        if (drawBelowNowLine) {
            bottomCropTime = timeDrawingBottom;
        } else if (isScrolling) {
            if (timeNow > timeDrawingBottom) {
                // scrolled way above now line
                bottomCropTime = timeDrawingBottom;
            } else if (timeNow >= timeDrawingTop && timeNow <= timeDrawingBottom) {
                bottomCropTime = timeNow;
            } else {
                // scrolled so that no data is visible, just paint column borders and exit
                paintColumnBorders(g2Component, columnModel);
                return;
            }
        } else {
            bottomCropTime = timeNow;
        }

        if (advanceMode && USE_OPTIMIZED_PLAYMODE_DRAWING) {
            paintColumnBorders(g2Buffer, columnModel);

            paintAllEventData(g2Buffer, timeDrawingTop, timeDrawingBottom, timeDrawingBottom, 
                columnModel);

            Shape oldClip = g2Component.getClip();
            g2Component.setClip(0, 0, getDrawableWidth(), getYCoordForTime(bottomCropTime));
            g2Component.drawImage(bufferedImage, null, 0, 0);
            g2Component.setClip(oldClip);

            paintColumnBorders(g2Component, columnModel);

            paintNowLine(g2Component, timeNow);
        } else {
            paintColumnBorders(g2Component, columnModel);
            paintAllEventData(g2Component, timeDrawingTop, bottomCropTime, timeDrawingBottom, 
                columnModel);
            paintNowLine(g2Component, timeNow);
        }

        setScrollBarValue();

        // DEBUGGING STUFF - just uncomment what you need
        //System.out.println("-----------------------------------------");
        //System.out.println(paintCount+" advanceMode: "+advanceMode);
        //System.out.println(paintCount+" isScrolling: "+isScrolling);
        //System.out.println(paintCount+" isSteppingBackwards: "+isSteppingBackwards);
        //System.out.println(paintCount+" scrollBarValueSet: "+scrollBarValueSet);
        //System.out.println(paintCount+" isMouseDragging: "+isMouseDragging);
        //System.out.println(paintCount+" isMouseDraggingFinished: "+isMouseDraggingFinished);
        //System.out.println(paintCount+" currentDrawInterval: "+currentDrawInterval);
        //System.out.println(paintCount+" timeDrawingTop: "+timeDrawingTop);
        //System.out.println(paintCount+" timeDrawingBottom: "+timeDrawingBottom);
        //System.out.println(paintCount+" timeNow: "+timeNow);
        //System.out.println(paintCount+" bottomCropTime: "+bottomCropTime);
        //System.out.println(paintCount+" scrollbar value: "+jScrollBar1.getValue());
        //System.out.println(paintCount+" scrollbar extent: "+jScrollBar1.getVisibleAmount());
        //System.out.flush();

        if (isMouseDraggingFinished) {
            isMouseDraggingFinished = false;
            isMouseDragging = false;
        }
        
        prevDrawInterval = currentDrawInterval;
        previousDrawableWidth = getDrawableWidth();
        previousDrawableHeight = getDrawableHeight();
        paintCount++;
    }
    
    private final void paintAllEventData(
        Graphics2D g2, 
        float topTime, 
        float cropTime,
        float bottomTime,
        MSCColumnModel columnModel) 
    {
        // if nothing to draw, just return (otherwise would crash...)
        if (topTime > cropTime) return;
        
        List units = dataView.getUnitsForLayer(settings.getLayer(), timeDrawingTop, cropTime);

        Rectangle clipRectDA = new Rectangle(columnModel.getDrawingArea().getLeftEdge(), 0,
            columnModel.getDrawingArea().getWidth(), getYCoordForTime(cropTime));
        paintDrawingArea(g2, units, clipRectDA);

        Rectangle clipRectNC = new Rectangle(columnModel.getNotesCol().getLeftEdge(), 0, 
            columnModel.getNotesCol().getWidth(), getYCoordForTime(cropTime));
        Collection notes = mainFrame.getNoteManager()
            .getNotesInTimeRange(settings.getLayer(), timeDrawingTop, cropTime);
        paintNotesColumn(g2, columnModel, notes, clipRectNC);

        g2.setColor(COLOR_COLUMN_VALUES);
        paintColumnValues(g2, units, columnModel, getYCoordForTime(bottomTime), cropTime);
        
    }
    
    protected final void advanceDrawingTo(float oldTime, float newTime) {
        Graphics2D g2 = (Graphics2D)getGraphics();
        if (g2 == null) return;
        int rightEdge = getDrawableWidth()-1;
        int sliceTop = getYCoordForTime(oldTime);
        int sliceBottom = getYCoordForTime(newTime);
        g2.drawImage(bufferedImage, 
            0, sliceTop, rightEdge, sliceBottom,
            0, sliceTop, rightEdge, sliceBottom,
            null
        );
        
        paintNowLine(g2, newTime);
    }
    
    private final void paintNowLine(Graphics2D g2, float time) {
        g2.setColor(COLOR_NOWLINE);
        int y = getYCoordForTime(time);
        g2.drawLine(getInsets().left, y, getWidth()-getInsets().right-2, y);
    }

    private final void paintDrawingArea(Graphics2D g2, List units, Rectangle clipRect) {
        if (useAntiAlias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }

        Shape oldClip = null;
        if (!drawBelowNowLine) {
            oldClip = g2.getClip();
            g2.setClip(clipRect);
        }

        Stroke oldStroke = g2.getStroke();
        g2.setStroke(new BasicStroke(2.0f));
        paintEventLines(g2, units, clipRect);
        g2.setStroke(oldStroke);

        g2.setColor(COLOR_EVENT_VALUES);
        paintEventValues(g2, units, clipRect);

        if (!drawBelowNowLine) {
            g2.setClip(oldClip);
        }

        if (useAntiAlias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
        }
    }

    private final void paintEventLines(Graphics2D g2, List units, Rectangle area) {
        Iterator iter = units.iterator();
        while (iter.hasNext()) {
            TransferUnit element = (TransferUnit)iter.next();
            if (element.isOneUnit()) {
                if (element.isDropped()) {
                    g2.setColor(COLOR_EVENT_LINES);
                    drawDroppedUnit(g2, element, area);
                } else {
                    g2.setColor(COLOR_EVENT_LINES);
                    drawNormalUnit(g2, element, area);
                }
            } else {
                g2.setColor(COLOR_EVENT_AREA);
                drawFatUnit(g2, element, area);
            }
        }
    }

    private final void paintEventValues(Graphics2D g2, List units, Rectangle area) {
        Iterator iter = units.iterator();
        while (iter.hasNext()) {
            TransferUnit element = (TransferUnit)iter.next();
            drawUnitFieldValue(g2, element, area);
        }
    }

    /** Paints the event values inside the columns. Note that this method is
     * fully parameterized, it doesn't use any instance field values, only
     * values from parameters. This makes it flexible for drawing both directly on the
     * component or on a buffered image.
     *
     * @param g2 Graphics2D context where to draw.
     * @param units A list of TransferUnits whose values will be drawn.
     * @param columnModel The columnModel for the columns.
     * @param bottomCoordinate The absolute bottom coordinate for drawable area.
     * @param currentTime The imaginary 'timeNow' that will be used so that no
     * values for events happening after this time will be drawn. 
     */
    private final void paintColumnValues(
        Graphics2D g2, 
        List units, 
        MSCColumnModel columnModel,
        int bottomCoordinate,
        float currentTime)
    {
        Shape oldClip = g2.getClip();

        MSCColumnModel.MSCColumn[] colsLeft = columnModel.getLeftCols();
        MSCColumnModel.MSCColumn[] colsRight = columnModel.getRightCols();

        for (int i=0; i < colsLeft.length; i++) {
            Iterator iter = units.iterator();
            while (iter.hasNext()) {
                TransferUnit element = (TransferUnit)iter.next();

                if (element.getSource().equals(settings.getLeftHost())) {
                    paintSingleColumnValue(g2, colsLeft[i], element, currentTime, bottomCoordinate, true);
                    paintSingleColumnValue(g2, colsRight[i], element, currentTime, bottomCoordinate, false);
                } else if (element.getSource().equals(settings.getRightHost())) {
                    paintSingleColumnValue(g2, colsLeft[i], element, currentTime, bottomCoordinate, false);
                    paintSingleColumnValue(g2, colsRight[i], element, currentTime, bottomCoordinate, true);
                } else {
                    // hosts don't match, do nothing
                }
            } // end while

        } // end for

        g2.setClip(oldClip);

    } // end method paintColumns
    
    private final void paintSingleColumnValue(
        Graphics2D g2, 
        MSCColumnModel.MSCColumn column,
        TransferUnit unit,
        float currentTime,
        int bottomCoordinate,
        boolean isSending) 
    {
        float time;
        String value;
        if (isSending) {
            time = unit.getSendStart();
            value = unit.getValueSend(column.getVariable());
        } else {
            time = unit.getReceiveStart();
            value = unit.getValueReceive(column.getVariable());
        }
        
        if (value != null && (drawBelowNowLine || (time <= currentTime))) {
            // don't try to process more than necessary
            if (value.length() > MAX_DRAWABLE_VALUE_LENGTH) {
                value = value.substring(0, MAX_DRAWABLE_VALUE_LENGTH);
            }
            
            // get exact bounds for the values in this context
            // (needed for right-aligning the values)
            Rectangle2D bounds = g2.getFontMetrics().getStringBounds(value, g2);

            g2.setClip(column.getLeftEdge(), 0, 
                column.getWidth(), bottomCoordinate);

            
            int startX;
            if (bounds.getWidth() > column.getWidth()) {
                // if value doesn't fit completely, align beginning to left edge
                startX = column.getLeftEdge();
            } else {
                // if fits, align end of string with right edge
				startX = column.getRightEdge() - (int)bounds.getWidth();
            }
            
            // Draw value inside the column so that for sender 
            // the top edge of the visible string is aligned with event time and for 
            // receiver the bottom edge of string is aligned with event time.
            int startY;
            if (isSending) {
                startY = getYCoordForTime(time) + (int)bounds.getHeight();
            } else {
                startY = getYCoordForTime(time);
            }
            
            g2.drawString(value, startX, startY);
        }
        
    }

    private final void paintColumnBorders(Graphics2D g2, MSCColumnModel columnModel) {
        g2.setColor(COLOR_COLUMN_BORDERS);
        paintBordersForColumn(g2, columnModel.getNotesCol());
        for (int i=0; i < columnModel.getLeftCols().length; i++) {
            paintBordersForColumn(g2, columnModel.getLeftCols()[i]);
        }
        paintBordersForColumn(g2, columnModel.getDrawingArea());
        for (int i=0; i < columnModel.getRightCols().length; i++) {
            paintBordersForColumn(g2, columnModel.getRightCols()[i]);
        }
    }
    
    private final void paintBordersForColumn(Graphics2D g2, MSCPanel.MSCColumnModel.MSCColumn column) {
        g2.drawLine(column.getLeftEdge()-1, getInsets().top,
        column.getLeftEdge()-1, getHeight() - getInsets().bottom);
        g2.drawLine(column.getRightEdge()+1, getInsets().top,
        column.getRightEdge()+1, getHeight() - getInsets().bottom);
    }

    private final void paintNotesColumn(Graphics2D g2, MSCPanel.MSCColumnModel columnModel, 
        Collection notes, Rectangle area) 
    {
        Iterator iterator = notes.iterator();
        while (iterator.hasNext()) {
            Note element = (Note)iterator.next();
            g2.drawImage(noteImage, null, area.x + 2, getYCoordForTime(element.getTime()));
        }
    }

    private final void drawNormalUnit(Graphics2D g2, TransferUnit unit, Rectangle area) {
        int leftY, rightY;
        if (unit.getSource().equals(settings.getLeftHost())) {
            leftY = getYCoordForTime(unit.getSendStart());
            rightY = getYCoordForTime(unit.getReceiveEnd());
        } else {
            leftY = getYCoordForTime(unit.getReceiveEnd());
            rightY = getYCoordForTime(unit.getSendStart());
        }

        g2.drawLine(area.x, leftY, area.x + area.width, rightY);
    }

    private final void drawFatUnit(Graphics2D g2, TransferUnit unit, Rectangle area) {
        if (DISABLE_ANTIALIAS_FOR_FAT_UNITS) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
        }

        int leftYStart, leftYEnd, rightYStart, rightYEnd;

        if (unit.getSource().equals(settings.getLeftHost())) {
            leftYStart = getYCoordForTime(unit.getSendStart());
            leftYEnd = getYCoordForTime(unit.getSendEnd());
            rightYStart = getYCoordForTime(unit.getReceiveStart());
            rightYEnd = getYCoordForTime(unit.getReceiveEnd());
        } else {
            leftYStart = getYCoordForTime(unit.getReceiveStart());
            leftYEnd = getYCoordForTime(unit.getReceiveEnd());
            rightYStart = getYCoordForTime(unit.getSendStart());
            rightYEnd = getYCoordForTime(unit.getSendEnd());
        }

        int[] xPoints = new int[5];
        int[] yPoints = new int[5];
        xPoints[0] = area.x; yPoints[0] = leftYStart;
        xPoints[1] = area.x + area.width; yPoints[1] = rightYStart;
        xPoints[2] = area.x + area.width; yPoints[2] = rightYEnd;
        xPoints[3] = area.x; yPoints[3] = leftYEnd;
        xPoints[4] = area.x; yPoints[4] = leftYStart;

        g2.fillPolygon(xPoints, yPoints, 5);

        if (useAntiAlias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }
    }

    private final void drawDroppedUnit(Graphics2D g2, TransferUnit unit, Rectangle area) {
        int center = area.x + (area.width / 2);

        if (unit.getSource().equals(settings.getLeftHost())) {
            g2.drawLine(area.x, getYCoordForTime(unit.getSendStart()),
            center, getYCoordForTime(unit.getReceiveStart()));
        } else {
            g2.drawLine(area.x + area.width, getYCoordForTime(unit.getSendStart()),
            center, getYCoordForTime(unit.getReceiveStart()));
        }
        drawDroppedCross(g2, center, getYCoordForTime(unit.getReceiveStart()),
            DROPPED_CROSS_WIDTH);
    }

    private final void drawDroppedCross(Graphics2D g2, int x, int y, int width) {
        Color oldColor = g2.getColor();
        g2.setColor(COLOR_DROPPED_CROSS);
        g2.drawLine(x-width/2, y-width/2, x+width/2, y+width/2);
        g2.drawLine(x-width/2, y+width/2, x+width/2, y-width/2);
        g2.setColor(oldColor);
    }

    private final void drawUnitFieldValue(Graphics2D g2, TransferUnit unit, Rectangle area) {
        int leftY, rightY;
        if (unit.getSource().equals(settings.getLeftHost())) {
            leftY = getYCoordForTime(unit.getSendStart());
            rightY = getYCoordForTime(unit.getReceiveStart());
        } else {
            leftY = getYCoordForTime(unit.getReceiveStart());
            rightY = getYCoordForTime(unit.getSendStart());
        }

        double yDiff = rightY-leftY;
        double xDiff;

        if (unit.isDropped()) {
            xDiff = area.width / 2.0;
        } else {
            xDiff = area.width;
        }

        double angle = Math.atan(yDiff / xDiff);

        StringBuffer sb = new StringBuffer();
        VariableDefinition[] variables = settings.getVariablesCenter();
        for (int i = 0; i < variables.length; i++) {
            if (i>0) sb.append("  ");
            sb.append(variables[i].getName()+"="+unit.getValue(variables[i]));
        }

        Rectangle2D bounds = g2.getFontMetrics().getStringBounds(sb.toString(), g2);

        double rotatedHalfWidth = Math.abs(Math.cos(angle) * bounds.getWidth() / 2.0);
        
        int stringLeftEdge = (int)(area.x + area.width/2 - rotatedHalfWidth);
        if (stringLeftEdge < area.x) stringLeftEdge = area.x;

        int stringStartY;
        if (unit.isDropped()) {
            int lowerEnd = Math.max(leftY, rightY);
            int stringStartYCorrection = (int)(Math.tan(angle) * rotatedHalfWidth);
            stringStartY = lowerEnd - stringStartYCorrection - EVENT_VALUE_YDIFF;
        } else {
            int stringStartYCorrection = (int)(Math.tan(angle) * (double)(stringLeftEdge-area.x));
            stringStartY = leftY + stringStartYCorrection - EVENT_VALUE_YDIFF;
        }

        AffineTransform oldTransform = g2.getTransform();
        g2.transform(AffineTransform.getRotateInstance(angle, stringLeftEdge, stringStartY));

        g2.drawString(sb.toString(), stringLeftEdge, stringStartY);

        g2.setTransform(oldTransform);
    }

    private final int getYCoordForTime(float time) {
        return (int)((time - timeDrawingTop) * settings.getVisualScale());
    }
    
    private final float getTimeForYCoord(int y) {
        return ((float)y / (float)settings.getVisualScale()) + timeDrawingTop;
    }
    
    private final void doMouseSelecting(int y) {
        float time = getTimeForYCoord(y);
        if (time < timeDrawingTop + verticalMarginTime) {
             time = timeDrawingTop + verticalMarginTime;
        } else if (time > timeDrawingBottom - verticalMarginTime) {
            time = timeDrawingBottom - verticalMarginTime;
        }
        
        mainFrame.getAnimationTimeState().setNowTime(time);
    }
    
    public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt) {
        jScrollBar1.setValue(jScrollBar1.getValue() + 
            (evt.getUnitsToScroll() * jScrollBar1.getUnitIncrement()));
    }
    
} // end inner class MSCDrawingPanel


/** MSCColumnModel encapsulates the information about different columns drawn in
 * the MSCDrawingPanel. MSCColumnModel has four main parts, from left to right:
 * note column, left variable columns, drawing area, right variable columns.
 * Note column and drawing area are always represented by exactly one MSCColumn.
 * Left and right variable columns are represented by 0..n MSCColumn objects
 * each.
 *
 * <p>Currently MSCColumnModel is implemented so that the same columns always
 * exist in left and in right, and the order of columns is 'centric' in
 * relation to the drawing are: the same column is always closest to the
 * drawing area both in left and in right.
 * 
 * <p>The actual MSCColumn objects always represent the usable area between
 * the column border lines: the getLeftEdge returns the first pixel that can
 * be used for plotting and getRightEdge returns the last. MSCColumnModel takes
 * care of reserving one pixel between each column (and on the left and right
 * sides of the left- and rightmost columns).
 *
 * <p>Width is distributed between columns according to 
 * DataView.getValueMaxLength(VariableDefinition) for each variable, multiplied
 * by MSCPanel.WIDTH_COLUMN_PER_CHARACTER and the MSCPanel.WIDTH_COLUMN_NOTES
 * and MSCPanel.WIDTH_DRAWING_AREA_MINIMUM constants. The note column width
 * is always fixed. The variable columns never take more width than the
 * getValueMaxLength*WIDTH_COLUMN_PER_CHARACTER, so if there is extra width
 * after distributing that width to all columns, the extra width is used for
 * drawing area. However, if the getValueMaxLength*WIDTH_COLUMN_PER_CHARACTER
 * width is not available in full (after taking out WIDTH_COLUMN_NOTES and
 * WIDTH_DRAWING_AREA_MINIMUM plus 2 pixels for column borders), the width
 * is distributed so the drawing area gets WIDTH_DRAWING_AREA_MINIMUM and the
 * variable columns get what is left, relative to their own getValueMaxLength.
 * If the total width available for MSCColumnModel is less than
 * WIDTH_DRAWING_AREA_MINIMUM+WIDTH_COLUMN_NOTES+2, the total width is exceeded
 * when distributing the width for columns (it is recommended that this is
 * never allowed to happen in UI). Otherwise, as long as total width allows
 * for allocating width according to these constants, the model guarantees that
 * the right edge of the rightmost column is always at totalWidth-2.
 */
public class MSCColumnModel {

    private final MSCColumn columnNotes;

    /* Index 0 is closest to drawing area, i.e. right-most of left columns */
    private final MSCColumn[] columnsLeft;
    private final MSCColumn columnDrawingArea;

    /* Index 0 is closest to drawing area, i.e. left-most of right columns */
    private final MSCColumn[] columnsRight;

    private final int totalWidth;

    public MSCColumnModel(int width, VariableDefinition[] columnVariables) {
        this.totalWidth = width;
        int columnDemand = 0;

        for (int i=0; i < columnVariables.length; i++) {
            columnDemand += 2 * (1 + WIDTH_COLUMN_PER_CHARACTER *
                dataView.getValueMaxLength(columnVariables[i]));
        }

        float columnWidthFactor;
        int drawingAreaWidth;

        if (columnDemand > 0 &&
            WIDTH_COLUMN_NOTES + 2 + WIDTH_DRAWING_AREA_MINIMUM + columnDemand > width)
        {
            int availableForColumns = width - WIDTH_DRAWING_AREA_MINIMUM - WIDTH_COLUMN_NOTES - 2;
            columnWidthFactor = (float)availableForColumns / (float)columnDemand;
            drawingAreaWidth = WIDTH_DRAWING_AREA_MINIMUM;
        } else {
            columnWidthFactor = 1.0f;
            drawingAreaWidth = width - WIDTH_COLUMN_NOTES - columnDemand - 2;
        }

        columnNotes = new MSCColumn(null, 1, WIDTH_COLUMN_NOTES-1);

        float previousRightEdge = WIDTH_COLUMN_NOTES-1;

        columnsLeft = new MSCColumn[columnVariables.length];
        for (int i = columnVariables.length-1; i >= 0; i--) {
            float currentColumnWidth = dataView.getValueMaxLength(columnVariables[i])
                * WIDTH_COLUMN_PER_CHARACTER * columnWidthFactor;

            float newLeftEdge = previousRightEdge + 2f;
            float newRightEdge = previousRightEdge + 2f + currentColumnWidth - 1f;

            MSCColumn tempColumn = new MSCColumn(
                columnVariables[i],
                Math.round(newLeftEdge),
                Math.round(newRightEdge)
            );

            columnsLeft[i] = tempColumn;
            previousRightEdge = newRightEdge;
        }

        columnDrawingArea = new MSCColumn(
            null,
            Math.round(previousRightEdge + 2f),
            Math.round(previousRightEdge + 2f + drawingAreaWidth-1f)
        );

        previousRightEdge = previousRightEdge + 2f + drawingAreaWidth - 1f;

        columnsRight = new MSCColumn[columnVariables.length];
        for (int i=0; i < columnVariables.length; i++) {
            float currentColumnWidth = dataView.getValueMaxLength(columnVariables[i])
                * WIDTH_COLUMN_PER_CHARACTER * columnWidthFactor;

            float newLeftEdge = previousRightEdge + 2f;
            float newRightEdge = previousRightEdge + 2f + currentColumnWidth - 1f;
            
            /* Final correction for the rightmost edge of the whole column model.
             * Because of precision issues we will never get this right without
             * a manual correction... */
            if (newLeftEdge < width && Math.round(newRightEdge) > width-2) {
                newRightEdge = width-2;
            }

            MSCColumn tempColumn = new MSCColumn(
                columnVariables[i],
                Math.round(newLeftEdge),
                Math.round(newRightEdge)
            );

            columnsRight[i] = tempColumn;
            previousRightEdge = newRightEdge;
        }

    }

    /** Getter for property columnNotes.
     * @return Value of property columnNotes.
     *
     */
    public MSCColumnModel.MSCColumn getNotesCol() {
        return columnNotes;
    }

    /** Getter for property columnsLeft.
     * @return Value of property columnsLeft.
     *
     */
    public MSCColumnModel.MSCColumn[] getLeftCols() {
        return this.columnsLeft;
    }

    /** Getter for property columnDrawingArea.
     * @return Value of property columnDrawingArea.
     */
    public MSCColumnModel.MSCColumn getDrawingArea() {
        return columnDrawingArea;
    }

    /** Getter for property columnsRight.
     * @return Value of property columnsRight.
     */
    public MSCColumnModel.MSCColumn[] getRightCols() {
        return this.columnsRight;
    }

    public int getTotalWidth() {
        return totalWidth;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer("MSCColumnModel (total "+getTotalWidth()+")");

        if (columnNotes != null) {
            sb.append(", notes "+columnNotes.toString());
        }
        if (columnsLeft.length > 0) {
            sb.append(", left ");
            for (int i=0; i < columnsLeft.length; i++) {
                sb.append(columnsLeft[i].toString());
            }
        }

        if (columnDrawingArea != null) {
            sb.append(", center "+columnDrawingArea.toString());
        }

        if (columnsRight.length > 0) {
            sb.append(", right ");
            for (int i=0; i < columnsRight.length; i++) {
                sb.append(columnsRight[i].toString());
            }

        }

        return sb.toString();
    }


    /* A simple class to express the left and right edges of one column in the column model,
     * including also the reference to VariableDefinition in case of variable columns.
     * The left and right edge values are both included in the drawable inside area of the
     * column.
     */
    public class MSCColumn {

        private final VariableDefinition variable;
        private final int leftEdge;
        private final int rightEdge;

        public MSCColumn(VariableDefinition variable, int leftEdge, int rightEdge) {
            this.variable = variable;
            this.leftEdge = leftEdge;
            this.rightEdge = rightEdge;
        }

        public VariableDefinition getVariable() {
            return variable;
        }

        public int getLeftEdge() {
            return leftEdge;
        }

        public int getRightEdge() {
            return rightEdge;
        }

        public int getWidth() {
            return rightEdge - leftEdge + 1;
        }

        public String toString() {
            return "("+leftEdge+","+rightEdge+")";
        }
    }
}


private class HostVariableTableModel extends javax.swing.table.AbstractTableModel {

    private String[][] tableData;

    public HostVariableTableModel(Host host) {
        List staticVariables = host.getStaticVariables();
        tableData = new String[2][staticVariables.size()];
        int rowIndex = 0;

        Iterator iter = staticVariables.iterator();
        while (iter.hasNext()) {
            StaticVariable element = (StaticVariable)iter.next();
            tableData[0][rowIndex] = element.getName();
            tableData[1][rowIndex] = element.getValue();
            rowIndex++;
        }
    }

    public int getColumnCount() {
        return 2;
    }

    public int getRowCount() {
        return tableData[0].length;
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        return tableData[columnIndex][rowIndex];
    }

}

private class ColumnHeadingPanel extends JPanel {

    public ColumnHeadingPanel() {
        setOpaque(false);
    }

    protected void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        MSCPanel.MSCColumnModel columnModel = getColumnModel();

        g2.setColor(COLOR_COLUMN_BORDERS);
        paintBorders(g2, columnModel.getNotesCol());
        for (int i=0; i < columnModel.getLeftCols().length; i++) {
            paintBorders(g2, columnModel.getLeftCols()[i]);
        }

        paintBorders(g2, columnModel.getDrawingArea());
        for (int i=0; i < columnModel.getRightCols().length; i++) {
            paintBorders(g2, columnModel.getRightCols()[i]);
        }

        g2.drawLine(getInsets().left, getHeight() - getInsets().bottom - 1,
            getWidth() - getInsets().right, getHeight() - getInsets().bottom - 1);

        g2.setColor(COLOR_COLUMN_VALUES);

        paintLabels(g2, columnModel.getLeftCols());
        paintLabels(g2, columnModel.getRightCols());
    }

    private void paintLabels(Graphics2D g2, MSCPanel.MSCColumnModel.MSCColumn[] columns) {
        for (int i=0; i < columns.length; i++) {
            g2.setClip(columns[i].getLeftEdge(), 0,
            columns[i].getWidth(), getHeight());
            Rectangle2D bounds = g2.getFontMetrics().getStringBounds(
            columns[i].getVariable().getName(), g2);
            g2.drawString(columns[i].getVariable().getName(), columns[i].getLeftEdge(),
            (int)bounds.getHeight());
        }
    }

    private void paintBorders(Graphics2D g2, MSCPanel.MSCColumnModel.MSCColumn column) {
        g2.drawLine(column.getLeftEdge()-1, getInsets().top,
        column.getLeftEdge()-1, getHeight() - getInsets().bottom);
        g2.drawLine(column.getRightEdge()+1, getInsets().top,
        column.getRightEdge()+1, getHeight() - getInsets().bottom);
    }
}
    
}
