package fi.helsinki.dacopan.ui;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.Reader;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.prefs.Preferences;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.animseq.AnimationSequence;
import fi.helsinki.dacopan.contsig.AnimationTimeState;
import fi.helsinki.dacopan.contsig.ControlSignalsListener;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.DefaultDataView;
import fi.helsinki.dacopan.model.ENCTreeModel;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.NoteManager;
import fi.helsinki.dacopan.model.ScenarioStepIterator;
import fi.helsinki.dacopan.model.StepIterator;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;
import fi.helsinki.dacopan.pef.XMLProtocolEventsReader;
import fi.helsinki.dacopan.scenario.ScenarioFile;
import fi.helsinki.dacopan.settings.GeneralSettings;

/**
 * MainFrame controls the program execution. 
 */
public class MainFrame 
    extends JFrame 
    implements ControlSignalsListener
{
    /*
     * CONSTANTS
     */    
    public static final String DACOPAN_EXTENSION_PEF = "pef";
    public static final String DACOPAN_EXTENSION_SCENARIO = "sce";
    public static final int DACOPAN_STEP_BACKWARD = -1;
    public static final int DACOPAN_STEP_FORWARD = 1;

    // Animation mode constants
    public static final int DACOPAN_NO_MODE = 0;
    public static final int DACOPAN_EXPLORE_MODE = 1;
    public static final int DACOPAN_SCENARIO_MODE = 2;
    
    // Animator state constants
    private static final int DACOPAN_STATE_NO_FILE = 0;
    private static final int DACOPAN_STATE_PLAYING_ENC = 1;
    private static final int DACOPAN_STATE_PLAYING_MSC = 2;
    
    // key for the last directory used in a file chooser in the preferences
    private static final String DACOPAN_LAST_DIRECTORY = "last_directory";
        
    /*
     * FIELDS
     */
    
    // Data and settings
    private DataView dataView;
    private NoteManager noteManager;
    private GeneralSettings animatorSettings;
    private SettingsMSC currentMSCSettings = null;
    
    // User interface
    private UserInterface mainUI;
   
    // The animator state and mode 
    private int animatorMode = DACOPAN_NO_MODE;
    private int animatorState;    
    private FileInformation fileInMemory;   // Information about the file in memory
        
    // Control objects
    private AnimationTimeState animationTimeState = null;     
    private AnimationSequence animationSequence = new AnimationSequence(this);

    // Instances of the possible AbstractAnimationPanels
    private AbstractAnimationPanel ufoPanel;
    private AbstractAnimationPanel mscPanel;
    private AbstractAnimationPanel encPanel;
    private AbstractAnimationPanel notePanel;

    private AbstractAnimationPanel timePanel;
    private Layer currentLayer;
	private TransferUnit currentEncUnit;

    /**
     * This constructor can be used in testing the animator, as it works with a given 
     * dataview as if a PEF had already been loaded in the memory.
     */
    public MainFrame(DataView data, NoteManager noteManager) {
        this.dataView = data;
        this.noteManager = noteManager;        

        mainUI = new UserInterface(this);

        // We simulate a situation that a protocol events file is loaded
        // in the memory - but no scenario information yet
        fileInMemory = new FileInformation();        
        fileInMemory.useGeneratedTestData();
        
        // initialize animation panels (normally done only after loading a file)
        createDefaultSettings();
        currentLayer = animatorSettings.getDefaultLayer();
        setupAnimationModeMSC(animatorSettings.getSettingsMSC(currentLayer), true);
        setAnimatorMode(DACOPAN_EXPLORE_MODE);        
        
        mainUI.refresh();
    }

    
    /**
     * This is the normal constructor that just creates the MainFrame and sets up 
     * the animator, and then waits for the user to do something.
     */
    public MainFrame() {
        
        // Create empty instances of all possible scenario information
        dataView = new DefaultDataView();
        noteManager = new NoteManager();

        // Create the user interface
        mainUI = new UserInterface(this);

        // No file is loaded in the memory
        fileInMemory = new FileInformation();
        animatorState = DACOPAN_STATE_NO_FILE;
        setAnimatorMode(DACOPAN_NO_MODE);        
        
        // Draw the UI
        mainUI.refresh();
    }


    /*
     * USER ACTION RESPONSES
     * -------------------------------------------------------------
     * The following methods respond to action coming from the user
     * through the UI. They are all called from the UserInterface class.
     */
    
    /**
     * Starts or resumes the animation.
     * <p>
     * We just call animationTimeState to continue playing the 
     * MSC settings it has been set up to show.
     */
    public void playAnimation() {                     
        if(animationTimeState != null) {
            animationTimeState.play();
            animationSequence.recordStartMSC(
                animatorSettings.getSettingsMSC(currentLayer),
                animationTimeState.getNowTime());
        }
        mainUI.refresh();                               
    }
    
    /**
     * Pauses the animation   
     */
    public void pauseAnimation() {
        if(animationTimeState != null) {
            animationTimeState.pause();
        }
        mainUI.refresh();
    }

    /**
     * Rewinds the animation to the beginning
     */
    public void rewindAnimation() {
        if(animationTimeState != null) {
            animationTimeState.toBeginning();
            animationTimeState.pause();
        }
        mainUI.refresh();
    }
    
    /**
     * Fast forwards the animation to the end.
     */
    public void fastForwardAnimation() {
        if(animationTimeState != null) {
            animationTimeState.toEnd();
            animationTimeState.pause();
        }

        mainUI.refresh();        
    }
    
    /**
     * Steps in the animation. Can step both backward and forward.
     * 
     * @param stepDirection The direction for stepping
     */
    public void stepInAnimation(int stepDirection) {
        
        if(animationTimeState != null) {
            if(animationTimeState.isPaused()) {
                if (stepDirection < 0) {
                    animationTimeState.stepBackward();
                } 
                else if (stepDirection > 0) {
                    animationTimeState.stepForward();
                }
            }
        }
        
        mainUI.refresh();
    }
    
    /**
     * Sets the animation mode to the given mode: 
     * <ul>
     *  <li><code>DACOPAN_SCENARIO_MODE</code>: Scenario</li>
     *  <li><code>DACOPAN_EXPLORE_MODE</code>: Explore mode</li>
     * </ul>
     *  
     * @param mode
     * @throws IllegalArgumentException
     */
    public void setAnimatorMode(int mode)
        throws IllegalArgumentException 
    {
        if(mode == DACOPAN_SCENARIO_MODE) {            
            animatorMode = mode;
            animationSequence.setActive(true);            
        }
        else if(mode == DACOPAN_EXPLORE_MODE && !isInExploreMode()) {
            this.animatorMode = mode;
            animationSequence.setActive(false);
            
            // Set up the MSC mode using general default settings
            setupAnimationModeMSC(animatorSettings.getSettingsMSC(currentLayer), false);
        }
        else if(mode == DACOPAN_NO_MODE) {
    	   animatorMode = DACOPAN_NO_MODE;           
        }
        
        mainUI.setAnimatorModeInfo(animatorMode);
    }
    
    /**
     * Loads an animation file to the animator. The file can be either a protocol events
     * file or a scenario file.
     * 
     * @param file  The file to be loaded
     */
    public void loadAnimationFile(File file) {
            
        if(file != null) {
            
            // Check the file type and select the appropriate load method
            // based on that
            String extension = null;
            String name = file.getName();
            int i = name.lastIndexOf('.');

            if (i > 0 && i < name.length() - 1) {
                extension = name.substring(i + 1).toLowerCase();
            }

            if (extension != null) {
                DialogProgressIndicator indicator = new DialogProgressIndicator(
                        this, Localization.getString("panel.loading.title"));
                indicator.setMessage(Localization.getString("panel.loading.message"));
	            if (extension.equals(DACOPAN_EXTENSION_PEF)) {
	                FileLoader loader = new PefFileLoader(file, indicator);
	                loader.start();
	            }
	            else if (extension.equals(DACOPAN_EXTENSION_SCENARIO)) {
	                FileLoader loader = new ScenarioFileLoader(file, indicator);
	                loader.start();
	            }
	            indicator.show();
            }
                                
        }         
    
        mainUI.refresh();
    }


    /**
     * Saves the scenario that is currently in memory. If the scenario is 
     * not saved already, a file chooser dialog is shown.
     */
    public void saveAnimationFile(boolean forceFileChooser) {

        // If no file is in memory, we cannot save anything
        if (fileInMemory.isLoaded() == false) {
           
            System.out.println("No animation in memory");
        } 
        else {
            // The file used for saving the scenario
            File saveFile = null;

            /*
             * If the file in memory isn't a scenario file, the user has to select the file 
             * to save. This happens when the user has loaded a protocol events file 
             * and then wants to save it.
             */
            if (!fileInMemory.isScenarioFile() || forceFileChooser) {
                JFileChooser fc = getFileChooser();

                // Add file filters to accept only the types of files we want to use
                fc.addChoosableFileFilter(new DaCoPAnFileFilter(DACOPAN_EXTENSION_SCENARIO, Localization.getString("mainframe.filetype_animseq"))); //$NON-NLS-1$
                boolean successfullSelection = true;
                
                do {
                    int returnValue = fc.showSaveDialog(this);
                    successfullSelection = true;
    
                    if (returnValue == JFileChooser.APPROVE_OPTION) {
                        saveFile = fc.getSelectedFile();
                        
                        if(!saveFile.getName().endsWith("sce")) {
                        	saveFile = new File(saveFile.getPath() + ".sce");
                        }
                        
                        if(saveFile.exists()) {
                            int n = JOptionPane.showConfirmDialog(
                                    this,
                                    Localization.getString("mainframe.msg_overwrite"),
                                    Localization.getString("mainframe.title_file_exists"),                                
                                    JOptionPane.YES_NO_OPTION);
                            if(n == JOptionPane.NO_OPTION){
                            	successfullSelection = false;
                            }
                        }
                    }
                } while(successfullSelection == false);
            }
            
            /*
             * If the loaded file is a scenario file, just use that file for saving 
             * instead of creating a new one.
             */
            else {
                saveFile = fileInMemory.getFile();
            }

            // Do the saving
            if(saveFile != null) {
                if (saveScenarioFile(saveFile)) {
                    fileInMemory.setFile(saveFile, FileInformation.SCENARIO_FILE);
                    setFileModified(false);
                }
            }
        }
        
        mainUI.refresh();
    }

    public void setFileModified(boolean value) {
        fileInMemory.setModified(value);
    }
    
    /**
     * Closes the animation and returns to the start state
     */
    public void closeAnimation() {
        animatorState = DACOPAN_STATE_NO_FILE;        
        // TODO make this change the panels...
    }
    
    /**
     * Opens a file chooser in the directory the user browsed the previous time.
     * Whenever the user changes the directory in the chooser the location is memorized.
     */
    static JFileChooser getFileChooser() {
        // retrieve preferences explicitly for the package fi.helsinki.dacopan.ui
        final Preferences prefs = Preferences.userNodeForPackage(fi.helsinki.dacopan.ui.MainFrame.class);
        String dir = prefs.get(DACOPAN_LAST_DIRECTORY, null);
        JFileChooser chooser = new JFileChooser(dir);
        // track any directory changes they user makes
        chooser.addPropertyChangeListener(JFileChooser.DIRECTORY_CHANGED_PROPERTY, new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                prefs.put(DACOPAN_LAST_DIRECTORY, ((File)evt.getNewValue()).getAbsolutePath());
            }
        });
        return chooser;
    }

    /**
     * Changes the layer to the one requested.
     * 
     * @param layer
     */
    public void changeLayer(Layer layer) 
        throws IllegalStateException, IllegalArgumentException
    {
        if(layer == null) {
        	throw new IllegalArgumentException("Cannot change to unspecified layer");
        }
        
        if (isInScenarioMode()) {
            // Signal the animation sequence that the layer has been changed
            // so that if it wants to save the layer change, it knows about it.
            // At the same time check if changing layer is allowed in the current
            // scenario phase.            
            if (!animationSequence.layerChanged()) {
                mainUI.refresh();
                return;
            }
        }
        
        currentLayer = layer;
        
        // Create new animation panels on the new layer
        setupAnimationModeMSC(animatorSettings.getSettingsMSC(layer), false);       
    }
    
    
    /**
     * Refreshes the MSC mode with updated settings from animatorSettings. 
     * This method is called after the settings panel has been closed.
     */
    public void refreshSettings() {
        setupAnimationModeMSC(animatorSettings.getSettingsMSC(currentLayer), false);
    }
    
    
    public void refreshScenarioSettings() {
        animationSequence.refreshWindow();
    }
    

    
    /**
     * Returns a reference to the data view so that for example the animation panels 
     * can use the data.
     * 
     * @return DataView
     */
    public DataView getDataView() {
        return dataView;
    }

    public Layer getCurrentLayer() {
        return currentLayer;
    }
    
    /**
     * Returns the current instance of the NoteManager being used to
     * store the notes of the currently loaded scenario
     * 
     * @return NoteManager
     */
    public NoteManager getNoteManager() {
        return this.noteManager;
    }
    
    public FileInformation getFileInformation() {
        return fileInMemory;
    }
    
    /**
     * Returns a reference to the general settings for the animator.
     * 
     * @return DataView
     */
    public GeneralSettings getSettings() {
        return animatorSettings;
    }
    
    /**
     * Returns a reference to the general settings for the animator.
     * 
     * @return DataView
     */
    public void setSettings(GeneralSettings generalSettings) {
        this.animatorSettings=generalSettings;
        
        // to avoid the situation where currentLayer doesn't correspond to any layer in
        // animatorSettings
        if (animatorSettings.getSettingsMSC(currentLayer) == null) {
            currentLayer = animatorSettings.getDefaultLayer();
            refreshSettings();
        }
        
        setFileModified(true);
    }
    
    public AnimationTimeState getAnimationTimeState() {
        return animationTimeState;
    }

    /**
     * 
     * @return true if in explore mode, otherwise false
     */
    public boolean isInExploreMode() {
        if(animatorMode == DACOPAN_EXPLORE_MODE){
            return true;
        }
        else return false;
    }
    
    /**
     * 
     * @return true if in scenario mode, otherwise false
     */
    public boolean isInScenarioMode() {
        if(animatorMode == DACOPAN_SCENARIO_MODE){
            return true;
        }
        else return false;
    }
        
    public boolean isInMSCMode() {
        if(animatorState == DACOPAN_STATE_PLAYING_MSC) {
        	return true;
        }
        else {
        	return false;
        }
    }
    
    public boolean isInENCMode() {
        if(animatorState == DACOPAN_STATE_PLAYING_ENC) {
            return true;
        }
        else {
            return false;
        }
    }
    
    
    /**
     * Returns the current enc unit if in ENC mode.
     */
    public TransferUnit getCurrentENCUnit()
        throws IllegalStateException 
    {
        if(animatorState == DACOPAN_STATE_PLAYING_ENC) {
            return currentEncUnit;
        }
        else {
            throw new IllegalStateException("ENC units should only be queried in enc mode");
        }
    }
        
    /*
     * ANIMATION CONTROL
     * 
     * The following methods are used to control the animation modes and
     * animation sequences in the animator
     */
    

    /**
     * This method sets up the MSC animation mode by creating the needed animation panels and 
     * giving them correct settings. It can be called by the MainFrame itself (for example 
     * setting a default mode once loaded an animation file) or by the ControlSignals Framework.
     * 
     * @param mscSettings
     *            Settings for the MSC animation
     */
    public void setupAnimationModeMSC(SettingsMSC mscSettings, boolean resetNowTime) {

        // Store the new MSC settings
        currentMSCSettings = mscSettings;
        currentLayer = mscSettings.getLayer();
        animatorState = DACOPAN_STATE_PLAYING_MSC;

        float currentNowTime;
        if (resetNowTime) {
            if (isInScenarioMode()) {
                currentNowTime = mscSettings.getStartTime();
            } 
            else {
                currentNowTime = 0f;
            }
        } 
        else {
            if (animationTimeState != null) {
                currentNowTime = animationTimeState.getNowTime();
            } 
            else {
                currentNowTime = 0f;
            }
        }
        
        // Create needed animation panels, i.e. the main animation panel, UFO and Notes
        mscPanel = new MSCPanel(this, mscSettings, dataView, currentNowTime);
        ufoPanel = new UFOPanel(this, dataView, mscSettings.getLayer());
        notePanel = new NotePanel(this, mscSettings.getLayer());
        timePanel = new TimePanel(this, currentNowTime);
        
        // Add them to the Frame
        mainUI.setUfoPanel(ufoPanel);
        mainUI.setNotePanel(notePanel);
        mainUI.setMainPanel(mscPanel);
        mainUI.setTimePanel(timePanel);
                
        // Create the animation time state and add the panels 
        // to listen it
        
        StepIterator combinedIterator = new ScenarioStepIterator(
            mscSettings.getLayer(),
            dataView,
            noteManager);
        
        if (animationTimeState != null) {
            animationTimeState.discard();
        }
        
        if (isInScenarioMode()) {
            animationTimeState = new AnimationTimeState(
                currentNowTime,
                dataView.getEndTime(),
                mscSettings.getTimeScale(), 
                animatorSettings.getRefreshDelay(),
                combinedIterator,
                animationSequence,
                mscSettings.getEndTime());
        } else {
            animationTimeState = new AnimationTimeState(
                currentNowTime,
                dataView.getEndTime(),
                mscSettings.getTimeScale(), 
                animatorSettings.getRefreshDelay(),
                combinedIterator);
        }
        animationTimeState.addControlSignalsListener(ufoPanel);
        animationTimeState.addControlSignalsListener(notePanel);
        animationTimeState.addControlSignalsListener(mscPanel);
        animationTimeState.addControlSignalsListener(timePanel);
        animationTimeState.addControlSignalsListener(this);
                        
        currentEncUnit = null;                
        
        mainUI.refresh();
    }

    /**
     * When the animation comes back from the ENC mode we need to handle that event
     * in different ways depending on the animator mode:
     * <p>
     * <ul>
     *  <li>in explore mode we just continue from where we paused the animation when
     *      going to ENC.</li>
     *  <li>in scenario mode we notify the animation sequence to continue</li>
     * </ul>
     */
    public void continueAnimationModeMSC()
    {
        animatorState = DACOPAN_STATE_PLAYING_MSC;

        if (isInScenarioMode()) {
            animationSequence.notifyExitFromEnc();
        }
        
        // Put back the current panels
        mainUI.setUfoPanel(ufoPanel);
        mainUI.setNotePanel(notePanel);
        mainUI.setMainPanel(mscPanel);
        mainUI.setTimePanel(timePanel);
                        
        timePanel.stepTo(animationTimeState.getNowTime());

        mainUI.refresh();        
    }   

    
    /**
     * Sets up the ENC animation mode. This can be called by the Animation sequence framework 
     * or some panel that has received information from the user asking to change to Encapsulation
     * animation. Such panel can be either a UFO panel or an MSC panel.
     * 
     * @param unit
     */
    public void setupAnimationModeENC(TransferUnit unit) {   
        
       animatorState = DACOPAN_STATE_PLAYING_ENC;
        
       // Pause the current animation timestate so that when we come back 
       // from ENC, we can continue from where we left off.
       if (animationTimeState != null) {
           animationTimeState.pause();
       }
       if(isInScenarioMode()) {
       	    animationSequence.recordEnc(unit);
       }
       
       
       // Create animation panels
       ENCTreeModel etm = new ENCTreeModel(unit, dataView);
       encPanel = new ENCPanel(this, etm);
       UFOPanel emptyUfoPanel = new UFOPanel(this, null); // Show an empty UFO panel
       NotePanel encNotePanel = new NotePanel(this, etm.getRootUnit());
       
       // TODO need to make sure that the note panel fully "disables"
       //itself if the layer param is null       
       
       // Set the mainUI to show the new panels
       mainUI.setUfoPanel(emptyUfoPanel);
       mainUI.setNotePanel(encNotePanel);
       mainUI.setMainPanel(encPanel);

       //also call mainUI to disable selected elements of toolbar
       //(everything except language and help)
       mainUI.refresh();
       currentEncUnit = unit;
    }

    /**
     * Returns a collection of objects that will be saved in a scenario file. This method is 
     * currently NOT used for loading.
     */
    private Collection getSaveables() {
        Collection saveables = new LinkedList();
        saveables.add(animationSequence);
        saveables.add(noteManager);
        saveables.add(animatorSettings);
        return saveables;
    }
    
 
    /**
     * Saves the scenario that is in the memory to the given file.
     * 
     * @param file
     * @return true, if saving was successfull. Otherwise false.
     */
    private boolean saveScenarioFile(File file) {

        ScenarioFile scenarioFile; 
        
        try {
            if (fileInMemory.isScenarioFile()) {
                scenarioFile = new ScenarioFile(file);
            } else {
                Reader pefReader = new BufferedReader(new FileReader(fileInMemory.getFile()));
                scenarioFile = new ScenarioFile(file, pefReader);
            }

            Collection saveables = getSaveables();

            scenarioFile.save(saveables);
            
            return true;
        } catch (Exception e) {
            JOptionPane.showMessageDialog(this, e.toString(), "Error while saving scenario file",
                JOptionPane.ERROR_MESSAGE);
            
            return false;
        }
    }

    /**
     * Checks if the given file is a scenario file or not. This method is used to determine if a 
     * file selection dialog should be shown or not when saving a file.
     * 
     * @param loadedFile
     * @return
     */
    private boolean isScenarioFile(File loadedFile) {

        // Check the file extension to see if it is a scenario file or not

        return false;
    }


    /**
     * Creates default GeneralSettings and MSC settings for all layers based 
     * on the data that is currently in DataView.
     */
    private void createDefaultSettings() {

        // Make sure there is a general settings object
        animatorSettings = new GeneralSettings();
        
        // Read the layers from DataView 
        List layers = dataView.getLayers();
        SettingsMSC settingsMSC = null;
        
        // Read the hosts from DataView
        List hosts = dataView.getHosts();
        if(hosts.size() < 2) {
            throw new RuntimeException("Must have two hosts!");
        }
        Host host1 = (Host)hosts.get(0);
        Host host2 = (Host)hosts.get(1);

        // Generate settings for each layer
        Iterator iter = layers.iterator();
        while (iter.hasNext()) {
            Layer element = (Layer) iter.next();
            // Generate the settings for this layer
            settingsMSC = new SettingsMSC(host1, host2, element,
            		new VariableDefinition[0], new VariableDefinition[0]);
            
            // Add the layer settings to GeneralSettings
            animatorSettings.addSettingsMSC(settingsMSC);
        }
        currentLayer = animatorSettings.getDefaultLayer();
    }
    

    
    

    /**
     * Creates the MainFrame object and shows it - and so it starts running :) 
     * 
     * @param args
     */
    public static void main(String[] args) {
        MainFrame mainFrame = new MainFrame();
        mainFrame.show();
    }
    
    // ControlSignalsListener methods
   
    public void advance(float step, float nowTime) {
    }
    
    public void stepTo(float nowTime) {
    }
    
    public void toPauseMode() {
        mainUI.refresh();
    }
    
    public void toPlayMode() {
        mainUI.refresh();
    }
    
    
    /**
     * @return
     */
    public boolean isPaused() {
        if(animationTimeState != null) {
            return animationTimeState.isPaused();
        }
        else {
            return true;
        }
    }
        
    /**
     * Tells whether the animation currently in memory is
     * at the beginning.
     */
    public boolean isAtBeginning() {    	
        return animationTimeState.isAtBeginning();
    }
        
    /**
     * Tells whether the animation currently in memory has
     * reached it's end.
     */
    public boolean isAtEnd() {        
        return animationTimeState.isAtEnd();
    }
    
    /**
     * @return
     */
    public SettingsMSC getCurrentMSCSettings() {
        return currentMSCSettings;
    }
    
    /**
     * @return
     */
    public int getMode() {
        return animatorMode;
    }
    
    public void invokeChangeSettings() {
        mainUI.changeSettings();
    }

	/**
	 * @return
	 */
	public boolean isAnimationSequenceDialogActive() {
		return animationSequence.isActive();
	}

	/**
	 * 
	 */
	public void showScenarioDialog() {        
		animationSequence.setActive(true);
        mainUI.refresh();
	}


	/**
	 * 
	 */
	public void refreshUI() {
		mainUI.refresh();		
	}

	/**
	 * Loads a file in a separate thread.
	 */
	private abstract class FileLoader extends SwingWorker {

	    protected DialogProgressIndicator indicator;
        protected DataView newDataView = new DefaultDataView();
	    protected File file;
	    protected Exception e;
	    
	    public FileLoader(File file, DialogProgressIndicator indicator) {
	        this.file = file;
	        this.indicator = indicator;
	    }
	    
	    public final Object construct() {
	        Reader fileReader = null;
	        try {
	            fileReader = new FileReader(file);
	            doConstruct(fileReader);
	        } catch(Exception e) {
	            this.e = e;
	        } finally {
	            if(fileReader != null) {
	                try { fileReader.close(); } catch(Exception ignored) { }
	            }
	        }
	        return null;
	    }
	    
        public void finished() {
            try {
                doFinished();
            } catch(Exception e) {
                String errorString = Localization.getString("mainframe.msg_invalid_pef") + e.getMessage(); //$NON-NLS-1$
                JOptionPane.showMessageDialog(MainFrame.this, errorString, 
                        Localization.getString("mainframe.title_loading_pef_failed"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
            } finally {
                indicator.close();
            }
        }
        
	    protected void handleException(Exception e) {
            e.printStackTrace();
            if(e instanceof FileNotFoundException) {
	            JOptionPane.showMessageDialog(MainFrame.this, Localization.getString("mainframe.msg_file_not_found") + e.getMessage(),  //$NON-NLS-1$
	                    Localization.getString("mainframe.title_loading_pef_failed"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
	        } 
            else {
	            String errorString = Localization.getString("mainframe.msg_invalid_pef") + e.getMessage(); //$NON-NLS-1$
	            JOptionPane.showMessageDialog(MainFrame.this, errorString, 
	                    Localization.getString("mainframe.title_loading_pef_failed"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
	        }
	    }
	    
	    protected abstract void doConstruct(Reader reader) throws Exception;

	    protected abstract void doFinished() throws Exception;
	    
	}
	
	/**
	 * Loads a PEF file into the memory in a separate thread.
	 */
	private final class PefFileLoader extends FileLoader {
        
	    public PefFileLoader(File file, DialogProgressIndicator indicator) {
            super(file, indicator);
        }

        protected void doConstruct(Reader fileReader) throws Exception {
            XMLProtocolEventsReader xmlReader = new XMLProtocolEventsReader();
            xmlReader.read(fileReader, newDataView, indicator);
        }
	    
        protected void doFinished() {
            
            if(e != null) {
                handleException(e);
                return;
            }
            
            // Save some information about the animator state
            fileInMemory.setFile(file, FileInformation.PROTOCOL_EVENTS_FILE);
            fileInMemory.setModified(true);
            
            // Now that the loading has been done successfully, replace
            // the data view.
            dataView = newDataView;
            noteManager = new NoteManager();
            
            // Remove any old animation sequence information
            animationSequence.setActive(false);
            animationSequence = new AnimationSequence(MainFrame.this);
            
            // Create a default SettingsMSC object based on
            // the data loaded from the PEF
            createDefaultSettings();
            
            currentLayer = animatorSettings.getDefaultLayer();
            setAnimatorMode(DACOPAN_EXPLORE_MODE);
            setupAnimationModeMSC(animatorSettings.getSettingsMSC(
                animatorSettings.getDefaultLayer()), true);
            
            indicator.setProgress(100);
            indicator.close();

            
        }

	}

	private final class ScenarioFileLoader extends FileLoader {

	    private AnimationSequence newAnimationSequence = new AnimationSequence(MainFrame.this);
	    private NoteManager newNoteManager = new NoteManager();
	    private GeneralSettings newSettings = new GeneralSettings();
	    
        public ScenarioFileLoader(File file, DialogProgressIndicator indicator) {
            super(file, indicator);
        }
        
        protected void doConstruct(Reader reader) throws Exception {
            
            Collection saveables = new LinkedList();
            saveables.add(newAnimationSequence);
            newNoteManager = new NoteManager();
            saveables.add(newNoteManager);
            newSettings = new GeneralSettings();
            saveables.add(newSettings);
            
            ScenarioFile scenarioFile = new ScenarioFile(file);
            scenarioFile.load(saveables, newDataView, indicator);
        }
        
        protected void doFinished() {
            
            if(e != null) {
                handleException(e);
                return;
            }
            
            animationSequence.setActive(false);
            animationSequence = newAnimationSequence;

            noteManager = newNoteManager;
            dataView = newDataView;
            animatorSettings = newSettings;
            setAnimatorMode(DACOPAN_SCENARIO_MODE);                                               
            
            fileInMemory.setFile(file, FileInformation.SCENARIO_FILE);
            fileInMemory.setModified(false);
            
            // If there is some items in the playlist, select the 
            // first of these. Otherwise start in the default MSC view
            createDefaultSettings();
            if(animationSequence.hasShowables()) {                            
                animationSequence.showFirstShowable();
                if (animationTimeState != null) {
                    animationTimeState.pause();
                }
            }
            else {
                setupAnimationModeMSC(newSettings.getSettingsMSC(
                    newSettings.getDefaultLayer()), true);
            }
            
            indicator.setProgress(100);
            indicator.close();
            
        }
	}
	
}