package fi.helsinki.dacopan.ui;

import java.awt.BorderLayout;
import java.awt.SystemColor;
import java.util.Iterator;
import java.util.List;

import javax.swing.JTabbedPane;
import javax.swing.JPanel;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.settings.*;


/**
 * <b>SettingsPanel </b> allows to change the animation settings.
 * <p>
 * The settings panel depend on the animation mode running and show a different
 * set of options that can be changed depending on that.
 * </p>
 * <p>  
 *  In the explore mode is in which more settings are shown to be tuned. These
 *  options are the following:
 *      <ul>
 *          <li>Scale options: allow to change values related with the scale used to
 *              represent the data contained in the protocol events file or scenario file.
 *          </li>
 *          <li>Performance options: change values that affect the performance of the
 *              animation and thus the animations can be run in slower machines
 *          </li>
 *          <li>Layer options: a panel per layer is shown to select the variables and
 *              protocol header fields that can be visualized in the animation
 *          </li>
 *      </ul>
 * </p>
 * <p>
 * In scenario mode only is shown information about the variables and protocol
 * event fields to be selected for display in the animation. Also the value for the 
 * visual scale for the active layer can be set.
 * </p>
 * 
 * @version $Id: SettingsPanel.java,v 1.25 2004/05/16 16:31:19 arrastia Exp $
 */
public class SettingsPanel extends javax.swing.JPanel {
   
    /**
     * Creates a new panel to be used in EXPLORE mode
     * 
     * @param dataView
     *            reference to a DataView object to access to some data contained in the data
     *            structures
     * @param generalSettings
     *            reference to the GeneralSettings object that stores the current settings
     */
    public SettingsPanel(DataView dataView, GeneralSettings generalSettings) {
        if (dataView == null || generalSettings == null) {
            throw new IllegalArgumentException("Null parameters not accepted!");
        }
        
        List layers;
        this.dataView=dataView;
        this.generalSettings = generalSettings;

        this.setLayout(new java.awt.BorderLayout());
        
        /* initialize the JTabPane that will contain the different settings panels */ 
        initSettingsJTabPane();

        /* initialize the Performance Settings and the Scale Settings to be used in all layers */
        initGeneralSettingsPanel();

        /* initialize the JPanels that contain the settings that can be modified in every layer */
        initAllLayersSettings();
    }

    /**
     * Creates a new panel to be used in SCENARIO mode
     * 
     * @param dataView
     *            reference to a DataView obkect to access to some data contained in the data
     *            structures
     * @param generalSettings
     *            reference to the GeneralSettings object that stores the current settings
     * @param layer
     *            current layer to be modified
     */
    public SettingsPanel(DataView dataView, GeneralSettings generalSettings, Layer layer) {
        
        if (dataView == null || generalSettings == null || layer == null) {
            throw new IllegalArgumentException("Null parameters not accepted!");
        }
        
        this.dataView = dataView;
        SettingsMSC settingsMSC = generalSettings.getSettingsMSC(layer);

        this.setLayout(new java.awt.BorderLayout());
        
        /* initialize the JTabPane that will contain the different settings panels */
        initSettingsJTabPane();

        /* create and add the settings panel used to change the scale settings of the layer
         * passed as parameter */
        this.scaleSettingsPanel = new SettingsPanelScale(generalSettings, layer);
        this.tabSettingsPanel.addTab(
                Localization.getString("panel.settings.tab_title_scale"),
                this.scaleSettingsPanel);
        
        /* create and add the settings panel for the layers */
        JPanel settingsPanelLayer = new SettingsPanelLayer(dataView, settingsMSC ,layer, true);
        this.tabSettingsPanel.addTab(layer.getName(), settingsPanelLayer); //$NON-NLS-1$
    }
    
    /**
     * Creates a new panel to be used in SCENARIO mode receiving as parameter a reference to a
     * SettingMSC object
     * 
     * @param dataView
     *            reference to a DataView object to access to some data contained in the data
     *            structures
     * @param settingsMSC
     *            SettingsMSC object to create the panel
     */
    public SettingsPanel(DataView dataView, SettingsMSC settingsMSC) {
        
        if (dataView==null || settingsMSC==null){
            throw new IllegalArgumentException("Null parameters not accepted!");
        }
        this.dataView = dataView;

        this.setLayout(new java.awt.BorderLayout());
        
        /* initialize the JTabPane that will contain the different settings panels */        
        initSettingsJTabPane();

        /* create and add the settings panel used to change the scale settings of the layer
         * passed as parameter */        
        this.scaleSettingsPanel = new SettingsPanelScale(settingsMSC);
        JPanel settingsPanelLayer = new SettingsPanelLayer(
            dataView, 
            settingsMSC,
            settingsMSC.getLayer(), 
            true);
        
        /* create and add the settings panel for the layers */
        this.tabSettingsPanel.addTab(Localization.getString("panel.settings.tab_title_scale"),this.scaleSettingsPanel); //$NON-NLS-1$
        this.tabSettingsPanel.addTab(
            settingsMSC.getLayer().getName(), 
            settingsPanelLayer); //$NON-NLS-1$
    }
    
    
    /**
     * Initializes the panels that are specific for the EXPLORE mode. <br>
     * These are basically the panels with the performance settings and scale settings that are
     * changed for all the layers.
     */
    private void initGeneralSettingsPanel(){
        
        /* initialize and add a Scale Settings Panel */
        this.scaleSettingsPanel = new SettingsPanelScale(generalSettings);
        this.tabSettingsPanel.addTab(
                Localization.getString("panel.settings.tab_title_scale"),
                this.scaleSettingsPanel); //$NON-NLS-1$
        
        /* initialize and add a Performance Settings Panel */        
        this.performanceSettingsPanel = new SettingsPanelPerformance(generalSettings);
        this.tabSettingsPanel.addTab(
                Localization.getString("panel.settings.tab_title_performance"),
                this.performanceSettingsPanel); //$NON-NLS-1$
    }
    
    /**
     * Creates as many SettingsPanelLayer as layers are. These panels are added to the
     * tabSettingsPanel. <br>This method is meant to be used in EXPLORE mode.
     */
    private void initAllLayersSettings(){
        List layers = this.dataView.getLayers();
        Layer auxLayer;
        SettingsPanelLayer auxSettingPanelLayer; 
        int index = 0;
        
        /* loop to create the panels and add them to the JTabPane */ 
        this.layerSettingsPanels = new SettingsPanelLayer[layers.size()];
        for (Iterator i = layers.iterator();i.hasNext();){
        	auxLayer=(Layer)i.next();
            this.layerSettingsPanels[index]=new SettingsPanelLayer(this.dataView, generalSettings.getSettingsMSC(auxLayer), auxLayer, false);
            this.tabSettingsPanel.addTab(auxLayer.getName(),this.layerSettingsPanels[index]);
            index++;                                    
        }
    }
    
    /** 
     * initializes the JTabPane that will contain the different settings panels. After that the JTabPane is added to the
     * class (extends JFRAME) by calling the add method for the panel
     */
    private void initSettingsJTabPane() {
        tabSettingsPanel = new JTabbedPane();
        
        tabSettingsPanel.setBackground(SystemColor.control);
        tabSettingsPanel.setForeground(SystemColor.textText);
        
        this.add(tabSettingsPanel, BorderLayout.CENTER);
    }
    
    /** Contains the different panels for the settings that can be accessd by changing the tabs*/
    private JTabbedPane tabSettingsPanel;

    /** JPanel with the settings for the scale of the animation. <br> 
     * Used in EXPLORE and SCENARIO modes.*/
    private SettingsPanelScale scaleSettingsPanel;

    /** JPanel with the settings for the performance of the animation. <br> 
     * Used in EXPLORE mode.*/
    private SettingsPanelPerformance performanceSettingsPanel;

    /** contains the panel with the settings for the settings that can be set for a layer. <br> 
     * Used in EXPLORE mode (several) and SCENARIO MODE (a single one).*/
    private SettingsPanelLayer [] layerSettingsPanels;

    /** GeneralSettings instance that will be used to access to the data that will be 
     * displayed in the different panels. */
    private GeneralSettings generalSettings;
    
    /** DataView instance used to access to some data stored in the data structures that
     * represent the connections */
    private final DataView dataView;
    
    // TODO maybe the attributes defined to store the panels are not necessary because they are
    // contained in the JTabPane. Make local variables and add those ones to the JTabPane
}
