package fi.helsinki.dacopan.ui;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.DefaultListCellRenderer;
import javax.swing.JList;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.VariableDefinition;



/**
 * <p>
 * JPanel used to change settings related to a specific layer.
 * </p>
 * <p>
 * With this panel the fields and header fields shown in the MSC animation are selected.
 * <br>This panel is shown in EXPLORE MODE for all the layers existing in the animation.
 * <br>In SCENARIO MODE the layer shown is the active layer.
 * <br>This panel is also created for a SettingsMSC object, allowing the edition of the settings
 * selected for a element in the scenario playlist. 
 * </p>
 * <p>
 * These settings are:
 * <ul>
 * <li>list of variables that will be shown in the animation (columns).</li>
 * <li>list of protocol header fields that will be shown in the animation (center).</li>
 * <li>autoplay for the scenario mode.</li>
 * </ul>
 * </p>
 * 
 * @version $Id: SettingsPanelLayer.java,v 1.26 2004/05/23 18:36:31 arrastia Exp $
 */
public class SettingsPanelLayer extends javax.swing.JPanel {
    
    /** 
     * Internal Class that contains information about how the JList should be rendered
     */
    class SettingsPanelCellRenderer extends DefaultListCellRenderer {

        public SettingsPanelCellRenderer() {
            super();
        }
        
        public Component getListCellRendererComponent(
            JList list,
            Object value,
            int index,
            boolean isSelected,
            boolean cellHasFocus)
        {
            super.getListCellRendererComponent(list,value,index,isSelected,cellHasFocus);
            super.setText(((VariableDefinition)value).getProtocol().getName()+"-"+((VariableDefinition)value).getFullName());
            return this;         
        }      
    }
    
    /**
     * Creates new form LayerSettingsPanel instance
     * 
     * @param dataView
     *            instace of the DataView interface used to access to the dataq structures
     * @param settingsMSC
     *            instace of SettingsMSC that contains the settings to be accessed and modified
     * @param layer
     *            working layer used to get all the available header fields and variables in it
     * @param scenarioMode
     *            true if the panel created is used in scenario mode.
     */
    public SettingsPanelLayer(DataView dataView, SettingsMSC settingsMSC, Layer layer, boolean scenarioMode){
        this.dataView=dataView; 
        this.settingsMSC=settingsMSC;
        this.scenarioMode=scenarioMode;
        
        // TODO replace and test layer for settingsMSC.getLayer()
        initAvailableLists(layer);
        initComponents();
        if (this.scenarioMode){
            initScenarioPausePanel();
        }
        
        initSelectedValues();
        
        calculateCommentLabelDimension();
        
        /* highlight selected index for the different lists */
        refreshSelectedIndicesJList(variablesList, selectedVariables);
        refreshSelectedIndicesJList(headerFieldsList, selectedHeaderFields);

        updateListsFeatures();
    }
    
    /**
     * initializes the available variables to be shown in the variables list
     * 
     * @param layer
     *            working layer to get information through it
     */
    private void initAvailableLists(Layer layer){
              
        this.availableVariables = new Vector();
        this.availableHeaderFields = new Vector();
        
        /* request the available VariableDefinition to the Dataview interface  */
        List variableDefinitionList = dataView.getAvailableVariables(layer);
        /* distribute among the different types of VariableDefinition */
        for (Iterator j = variableDefinitionList.iterator(); j.hasNext();) {
            VariableDefinition auxVar = (VariableDefinition) j.next();
            /* Protocols Header Field */
            if (auxVar.isProtocolField()) {
                availableHeaderFields.add(auxVar);
            /* Variables */
            } else if (auxVar.isUnitVariable()) {
                availableVariables.add(auxVar);
            }
        }
    }

    /**
     * Reads the values stored in settingsMSC to find the indices to be marked as selected in the
     * JLists
     */
    private void initSelectedValues(){
        VariableDefinition [] auxVars = settingsMSC.getVariablesColumns();
        VariableDefinition [] auxFields = settingsMSC.getVariablesCenter();
        this.selectedVariables = new ArrayList();
        this.selectedHeaderFields = new ArrayList();

        /* compare VariableDefinitions to set the indices in the variables JList*/ 
        for (int i = 0; i < auxVars.length; i++) {
            for (int j = 0; j < variablesList.getModel().getSize(); j++) {
                if (((VariableDefinition) variablesList.getModel().getElementAt(j)).equals(auxVars[i])) {
                    selectedVariables.add(new Integer(j));
                    break;
                }
            }
        }
        
        /* compare VariableDefinitions to set the indices in the protocol header fields JList*/
        for (int i = 0; i < auxFields.length; i++) {
            for (int j = 0; j < headerFieldsList.getModel().getSize(); j++) {
                if (((VariableDefinition) headerFieldsList.getModel().getElementAt(j)).equals(auxFields[i])) {
                    selectedHeaderFields.add(new Integer(j));
                    break;
                }
            }
        }
    }

    /** 
     * sets the render properties of the JLists contained in the panel.
     */
    private void updateListsFeatures(){
        this.variablesList.setCellRenderer(new SettingsPanelCellRenderer());
        this.headerFieldsList.setCellRenderer(new SettingsPanelCellRenderer());
    }

     
    /**
     * adds or deleted an item in the list of selected variables or header fields to be shown later
     * in the animation
     * 
     * @param evt
     *            MouseEvent
     * @param list
     *            JList in which the selection of index happened
     * @param listSelections
     *            list with the indices currently selected of the JList in which the selection
     *            happened
     */
    private void changeSelectedIndices(java.awt.event.MouseEvent evt, JList list, List listSelections){
        
        int index = list.locationToIndex(evt.getPoint());
        if (listSelections.contains(new Integer(index))) {
            listSelections.remove(new Integer(index));
        } else {
            listSelections.add(new Integer(index));
        }
    }
    
    /**
     * Updates the SettingsMSC object with the VariableDefinitions to be shown in the header fields
     * area.
     */
    private void updateHeaderFieldsSettingsMSC(){
        Object []aux = headerFieldsList.getSelectedValues();
        VariableDefinition []auxVars = new VariableDefinition[aux.length];
        
        for (int i=0;i<aux.length;i++){
            auxVars[i]=(VariableDefinition)aux[i];
        }
        
        settingsMSC.setVariablesCenter(auxVars);
    }
    
    /**
     * Updates the SettingsMSC object with the current VariableDefinitions to be shown in the
     * variables area.
     */
    private void updateVariablesSettingsMSC(){
        Object []aux = variablesList.getSelectedValues();
        VariableDefinition []auxVars = new VariableDefinition[aux.length];
        
        for (int i=0;i<aux.length;i++){
            auxVars[i]=(VariableDefinition)aux[i];
        }
        
        settingsMSC.setVariablesColumns(auxVars);
    }
    
    /**
     * initializes the components contained in the panel
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        java.awt.GridBagConstraints gridBagConstraints;

        panelLists = new javax.swing.JPanel();
        panelVariablesList = new javax.swing.JPanel();
        variablesLabel = new javax.swing.JLabel();
        jScrollPane1 = new javax.swing.JScrollPane();
        variablesList = new JList(availableVariables);
        variablesComment = new javax.swing.JLabel();
        panelHeaderFields = new javax.swing.JPanel();
        headerFieldsLabel = new javax.swing.JLabel();
        jScrollPane2 = new javax.swing.JScrollPane();
        headerFieldsList = new JList(availableHeaderFields);
        headerFieldsCommment = new javax.swing.JLabel();

        setLayout(new java.awt.BorderLayout());

        panelLists.setLayout(new java.awt.GridBagLayout());

        panelVariablesList.setLayout(new java.awt.GridBagLayout());

        variablesLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        variablesLabel.setText(Localization.getString("panel.settings.layer.variables"));
        variablesLabel.setToolTipText(Localization.getString("panel.settings.layer.variables_tooltip"));
        panelVariablesList.add(variablesLabel, new java.awt.GridBagConstraints());

        variablesList.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mousePressed(java.awt.event.MouseEvent evt) {
                variablesListMousePressed(evt);
            }
        });
        variablesList.addMouseMotionListener(new java.awt.event.MouseMotionAdapter() {
            public void mouseDragged(java.awt.event.MouseEvent evt) {
                variablesListMouseDragged(evt);
            }
        });

        jScrollPane1.setViewportView(variablesList);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        panelVariablesList.add(jScrollPane1, gridBagConstraints);

        variablesComment.setForeground(java.awt.Color.red);
        variablesComment.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        variablesComment.setText(" ");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        panelVariablesList.add(variablesComment, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.fill = java.awt.GridBagConstraints.VERTICAL;
        gridBagConstraints.weighty = 1.0;
        panelLists.add(panelVariablesList, gridBagConstraints);

        panelHeaderFields.setLayout(new java.awt.GridBagLayout());

        headerFieldsLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        headerFieldsLabel.setText(Localization.getString("panel.settings.layer.header_fields"));
        headerFieldsLabel.setToolTipText(Localization.getString("panel.settings.layer.header_fields_tooltip"));
        panelHeaderFields.add(headerFieldsLabel, new java.awt.GridBagConstraints());

        headerFieldsList.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mousePressed(java.awt.event.MouseEvent evt) {
                headerFieldsListMousePressed(evt);
            }
        });
        headerFieldsList.addMouseMotionListener(new java.awt.event.MouseMotionAdapter() {
            public void mouseDragged(java.awt.event.MouseEvent evt) {
                headerFieldsListMouseDragged(evt);
            }
        });

        jScrollPane2.setViewportView(headerFieldsList);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        panelHeaderFields.add(jScrollPane2, gridBagConstraints);

        headerFieldsCommment.setForeground(java.awt.Color.red);
        headerFieldsCommment.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        headerFieldsCommment.setText(" ");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        panelHeaderFields.add(headerFieldsCommment, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.VERTICAL;
        gridBagConstraints.weighty = 1.0;
        panelLists.add(panelHeaderFields, gridBagConstraints);

        add(panelLists, java.awt.BorderLayout.CENTER);

    }//GEN-END:initComponents

    
    /**
     * initializa the elements for a panel that will be added to the JPanel extended by the class
     * and containing a checkbox to set the autoplay setting for the scenario mode.
     */
    private void initScenarioPausePanel(){
        panelPauseScenarioMode = new javax.swing.JPanel();
        checkboxPauseScenarioMode = new javax.swing.JCheckBox();       

        panelPauseScenarioMode.setLayout(new java.awt.GridBagLayout());

        if (settingsMSC.getScenarioAutoPlay()){
            checkboxPauseScenarioMode.setSelected(true);
        }
        checkboxPauseScenarioMode.setText(Localization.getString("panel.settings.layer.scenario_autostart"));
        checkboxPauseScenarioMode.setToolTipText(Localization.getString("panel.settings.layer.scenario_autostart_tooltip"));
        checkboxPauseScenarioMode.setHorizontalTextPosition(javax.swing.SwingConstants.LEADING);
        checkboxPauseScenarioMode.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                checkboxPauseScenarioModeActionPerformed(evt);
            }
        });

        panelPauseScenarioMode.add(checkboxPauseScenarioMode, new java.awt.GridBagConstraints());

        add(panelPauseScenarioMode, java.awt.BorderLayout.SOUTH);        
    }
    
    private void headerFieldsListMouseDragged(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_headerFieldsListMouseDragged
        refreshSelectedIndicesJList(headerFieldsList, selectedHeaderFields);
    }//GEN-LAST:event_headerFieldsListMouseDragged

    private void variablesListMouseDragged(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_variablesListMouseDragged
        refreshSelectedIndicesJList(variablesList, selectedVariables);
    }//GEN-LAST:event_variablesListMouseDragged

    private void checkboxPauseScenarioModeActionPerformed(java.awt.event.ActionEvent evt) {
        settingsMSC.setScenarioAutoPlay(checkboxPauseScenarioMode.isSelected());
    }
    
    /** 
     * method called when the mouse is presed into the headerFieldsList JList
     * @param evt MouseEvent
     */
    private void headerFieldsListMousePressed(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_headerFieldsListMousePressed
        if (!checkPermittedSelectHeaderField(evt)){
            /* Undoes the selection of the mouse */
            refreshSelectedIndicesJList(headerFieldsList, selectedHeaderFields);
            return;
        }

        changeSelectedIndices(evt, headerFieldsList, selectedHeaderFields);
        if (selectedHeaderFields.size() >= SettingsMSC.MAX_SHOW_HEADER_FIELDS) {
            headerFieldsCommment.setText(Localization.getString("panel.settings.layer.header_fields_comment_limit_reached"));
        }

        /* Undoes the selection of the mouse */
        refreshSelectedIndicesJList(headerFieldsList, selectedHeaderFields);
        updateHeaderFieldsSettingsMSC();
    }//GEN-LAST:event_headerFieldsListMousePressed

    /** 
     * method called when the mouse is presed into the variablesList JList
     * @param evt
     */
    private void variablesListMousePressed(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_variablesListMousePressed
        
        if (!checkPermittedSelectVariable(evt)){
            /* Undoes the selection of the mouse */
            refreshSelectedIndicesJList(variablesList, selectedVariables);
            return;
        }
        
        changeSelectedIndices(evt, variablesList, selectedVariables);
        if (selectedVariables.size() >= SettingsMSC.MAX_SHOW_VARIABLES) {
            variablesComment.setText(Localization.getString("panel.settings.layer.variables_comment_limit_reached"));
        }
        
        /* Undoes the selection of the mouse */        
        refreshSelectedIndicesJList(variablesList, selectedVariables);
        updateVariablesSettingsMSC();
    }//GEN-LAST:event_variablesListMousePressed

    
    /**
     * Calculates and sets the size that the labels for the comments should have 
     */
    private void calculateCommentLabelDimension(){
        
        Graphics graphics = getGraphics();
        Rectangle2D bounds1 = variablesComment.getFontMetrics(variablesComment.getFont()).getStringBounds(Localization.getString("panel.settings.layer.variables_comment_limit_reached"), graphics);
        Rectangle2D bounds2 = headerFieldsCommment.getFontMetrics(headerFieldsCommment.getFont()).getStringBounds(Localization.getString("panel.settings.layer.header_fields_comment_limit_reached"), graphics);

        
        double width = Math.max(bounds1.getWidth(), bounds2.getWidth())+1;
        double height = Math.max(bounds1.getHeight(), bounds2.getHeight())+1;
        
        variablesComment.setMinimumSize(new java.awt.Dimension((int)width, (int)height));
        headerFieldsCommment.setMinimumSize(new java.awt.Dimension((int)width, (int)height));

        variablesComment.setPreferredSize(new java.awt.Dimension((int)width, (int)height));
        headerFieldsCommment.setPreferredSize(new java.awt.Dimension((int)width, (int)height));
        
    }
    
    /**
      * Checks if a variable that has been selected can be added to the list of selected variables
      * in case that no more variables could be selected sets a text in the comment label.
      * 
      * @param evt MouseEvent
      * 
      * @return true if the variable can be selected from the list  
      */
    private boolean checkPermittedSelectVariable(java.awt.event.MouseEvent evt) {
        int index = variablesList.locationToIndex(evt.getPoint());
        if (selectedVariables.contains(new Integer(index))) {
            variablesComment.setText(" ");
            return true;
        } else if (selectedVariables.size() < SettingsMSC.MAX_SHOW_VARIABLES) {
            return true;
        } 
        
        return false;
    }
    
     /**
      * Checks if a protocol header field that has been selected can be added to the list of
      * selected protocol header fields in case that no more protocol header fields could be
      * selected sets a text in the comment label.
      * 
      * @param evt
      *            MouseEvent
      * @return true if the variable can be selected from the list
      */
    private boolean checkPermittedSelectHeaderField(java.awt.event.MouseEvent evt) {
        int index = headerFieldsList.locationToIndex(evt.getPoint());
        if (selectedHeaderFields.contains(new Integer(index))) {
            headerFieldsCommment.setText(" ");
            return true;
        } else if (selectedHeaderFields.size() < SettingsMSC.MAX_SHOW_HEADER_FIELDS) {
            return true;
        }
        
        return false;
    }

    /**
     * refreshes the values selected in the JList
     * 
     * @param list
     *            JList where the selections can be made
     * @param selections
     *            selections already made in the list
     */
    private void refreshSelectedIndicesJList(JList list, List selections){
        int [] indices = new int[selections.size()];
        
        int j = 0;
        for (Iterator i=selections.iterator();i.hasNext();){
            indices[j]=((Integer)i.next()).intValue();
            j++;
        }
       
        list.setSelectedIndices(indices);
    }
    
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel headerFieldsCommment;
    private javax.swing.JLabel headerFieldsLabel;
    private javax.swing.JList headerFieldsList;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JScrollPane jScrollPane2;
    private javax.swing.JPanel panelHeaderFields;
    private javax.swing.JPanel panelLists;
    private javax.swing.JPanel panelVariablesList;
    private javax.swing.JLabel variablesComment;
    private javax.swing.JLabel variablesLabel;
    private javax.swing.JList variablesList;
    // End of variables declaration//GEN-END:variables
    private javax.swing.JPanel panelPauseScenarioMode;
    private javax.swing.JCheckBox checkboxPauseScenarioMode;
    
    private javax.swing.JPanel panelVisualScale;
    private javax.swing.JTextField visualScaleTextField;
    private javax.swing.JLabel visualScaleLabel;
    private javax.swing.JSlider visualScaleSlider;
    
    private SettingsMSC settingsMSC;
    private DataView dataView;
    
    // TODO maybe change the vector datatype and create another model for the JList
    /** available variables to be shown in the list of variables*/
    private Vector availableVariables;

    /** available protocol header fields to be shown in the list of protocol header fields*/
    private Vector availableHeaderFields;
    
    /** list with the variables selected to be shown in the animation */
    private List selectedVariables;
    
    /** list with the protocol header fields selected to be shown in the animation */
    private List selectedHeaderFields;

    /** true if the panel was created for scenario mode */
    private boolean scenarioMode;
}