package fi.helsinki.dacopan;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.DefaultDataView;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.Link;
import fi.helsinki.dacopan.model.Note;
import fi.helsinki.dacopan.model.NoteManager;
import fi.helsinki.dacopan.model.Protocol;
import fi.helsinki.dacopan.model.StaticVariable;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

import junit.framework.TestCase;

/**
 * Abstract base class for test cases of Animator.
 * 
 * <p>Provides methods for creating transfer units and static fields for the
 * common layers/protocols for convenience. Also contains specialized
 * assertions (such as {@link #assertListsHaveSameContent(String, List, List)})
 * that might come in handy.</p>
 * 
 * @version $Id: AbstractTestCase.java,v 1.17 2004/05/18 13:24:56 jbrown Exp $
 */
public abstract class AbstractTestCase extends TestCase {
    
    public static final Layer APPLICATION = new Layer(1, "Application", 20);
    public static final Layer TRANSPORT = new Layer(2, "Transport", 10);
    public static final Layer NETWORK = new Layer(3, "Network", 0);
    
    public static final Protocol HTTP = new Protocol(1, "HTTP", APPLICATION);
    public static final Protocol TCP = new Protocol(2, "TCP", TRANSPORT);
    public static final Protocol IPv4 = new Protocol(3, "IPv4", NETWORK);
    
    public static final Host HOST_A = new Host(1, "1.1.1.1", null);
    public static final Host HOST_B = new Host(2, "2.2.2.2", null);
    
    private static long idSequence = 0;
    
    /**
     * Provides unique ids for the units in test cases.
     */
    public static long getNextId() {
        return idSequence++;
    }
    
    /**
     * Creates a transfer unit on an arbitrary layer.
     */
    public static TransferUnit createUnit(float start, float end) {
        return new TransferUnit(getNextId(), Collections.EMPTY_MAP, IPv4, null, HOST_A, HOST_B, start, end, false);
    }
    
    /**
     * Creates a unit with the given children.
     */
    public static TransferUnit createUnit(List children) {
        return new TransferUnit(getNextId(), Collections.EMPTY_MAP, IPv4, null, HOST_A, HOST_B, children, false);
    }
    
    /**
     * Asserts that the given lists have the same content in the same order.
     * Provides more feedback than using assertEquals(first, second) would.
     */
    public static void assertListsHaveSameContent(String msg, List first, List second) {
        if(first.size() != second.size()) {
            fail(msg + " : lists are of different size (" + first.size() + "/" + second.size());
        }
        Iterator a = first.iterator();
        Iterator b = second.iterator();
        for(int i = 0; a.hasNext() && b.hasNext(); i++) {
            assertEquals(msg + " : list elements at index " + i + " differ", a.next(), b.next());
        }
    }
    
    /**
     * Returns a DataView populated with the following network data:
     * 
     * Host: A (static variable X) and B (static variable Y)
     * Link: K (A <--> B) (static variable MTU)
     * Layers: network, transport, application
     * Protocols:
     *   ipv4  (variables: IP1, IP2)
     *   tcp   (variables: TCP1, TCP2)
     *   http  (variables: HTTP1, HTTP2)
     * 
     * Units that are structured like this:
     *               
     *             HTTP
     *          /        \
     *        TCP        TCP
     *       /   \       /  \
     *     IP    IP    IP    IP
     * 
     * You could get the previous tree by calling getRandomData(1, 2);
     * 
     * The direction of each unit is selected randomly.
     * 
     * @param appLayerUnits how many application (http) units to create
     * @param appToTransBranches how many branches of children inside each higher-level unit
     * @param netDropRatio ratio of dropped net units to non-dropped (i.e. 1.0f = all dropped, 0.0f = none dropped)
     * @param debug if true, dump created units to System.out
     * @return the populated view
     */
    public static DataView getRandomData(int appLayerUnits, int appToTransBranches, int transToNetBranches, float netDropRatio, boolean debug) {

        // network units sent every 100th ms
        final float unitInterval = 0.100f;
        // time of travel is randomly selected from 60..140ms
        final float travelTime = 0.060f;
        final float travelTimeVariance = 0.080f;
        
        // set up basic data
        DataView view = new DefaultDataView();
        Host a = new Host(1, "111.111.111.111", "firsthost");
        StaticVariable X = new StaticVariable("X", "ABCdef", a, null, null);
        a.getStaticVariables().add(X);
        Host b = new Host(2, "123.123.123.123", "secondhost");
        StaticVariable Y = new StaticVariable("Y", "defABC", b, null, null);
        b.getStaticVariables().add(Y);
        view.addHost(a);
        view.addHost(b);
        
        Link k = new Link(1, a, b);
        StaticVariable mtu = new StaticVariable("MTU", "1234", null, k, null);
        k.getStaticVariables().add(mtu);
        view.addLink(k);
        
        Layer network = new Layer(1, "network", 0);
        Layer transport = new Layer(2, "transport", 10);
        Layer application = new Layer(3, "application", 20);
        view.addLayer(network);
        view.addLayer(transport);
        view.addLayer(application);
        
        Protocol ip = new Protocol(1, "ipv4", network);
        ip.getVariables().add(new VariableDefinition("IP1", VariableDefinition.UNIT, ip));
        ip.getVariables().add(new VariableDefinition("IP2", VariableDefinition.UNIT_FIELD, ip));
        Protocol tcp = new Protocol(2, "tcp", transport);
        tcp.getVariables().add(new VariableDefinition("Variable1", VariableDefinition.UNIT, tcp));
        tcp.getVariables().add(new VariableDefinition("Variable2", VariableDefinition.UNIT, tcp));
        tcp.getVariables().add(new VariableDefinition("Variable3", VariableDefinition.UNIT, tcp));
        tcp.getVariables().add(new VariableDefinition("Variable4", VariableDefinition.UNIT, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField1", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField2", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField3", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField4", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField5", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField6", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("HeaderField7", VariableDefinition.UNIT_FIELD, tcp));
        tcp.getVariables().add(new VariableDefinition("TCP2", VariableDefinition.FLOW, tcp));
        Protocol http = new Protocol(3, "http", application);
        http.getVariables().add(new VariableDefinition("HTTP1", VariableDefinition.UNIT, http));
        http.getVariables().add(new VariableDefinition("HTTP2", VariableDefinition.UNIT_FIELD, http));
        view.addProtocol(ip);
        view.addProtocol(tcp);
        view.addProtocol(http);

        // generate unit data
        if(debug) {
            System.out.println("Generated data:");
        }
        long id = 0;
        float currentTime = 0.0f;
        for(int i = 0; i < appLayerUnits; i++) {
            // randomize direction of each app unit
            boolean toA = random.nextBoolean();
            Host from = toA ? b : a;
            Host to = toA ? a : b;
            List transportUnits = new LinkedList();
            for(int j = 0; j < appToTransBranches; j++) {
                List networkUnits = new LinkedList();
                for(int s = 0; s < transToNetBranches; s++) {
                    boolean dropped = random.nextFloat() < netDropRatio;
                    float endTime = currentTime + travelTime + (travelTimeVariance * random.nextFloat());
                    if(dropped) {
                        endTime -= (endTime - currentTime) / 2;
                    }
                    TransferUnit networkUnit = new TransferUnit(
                            ++id,
                            generateValues(ip),
                            ip,
                            null,
                            from,
                            to,
                            currentTime,
                            endTime,
                            dropped
                    );
                    networkUnits.add(networkUnit);
                    currentTime += unitInterval;
                }
                TransferUnit transportUnit = new TransferUnit(
                        ++id,
                        generateValues(tcp),
                        tcp,
                        null,
                        from,
                        to,
                        networkUnits,
                        false
                );
                transportUnits.add(transportUnit);
            }
            TransferUnit appUnit = new TransferUnit(
                    ++id,
                    generateValues(http),
                    http,
                    null,
                    from,
                    to,
                    transportUnits,
                    false
            );
            view.addTransferUnit(appUnit, true);
            if(debug) {
                dumpUnit(appUnit, "");
            }
        }
        
        // return view
        return view;
    }
    
    private static void dumpUnit(TransferUnit unit, String indent) {
        System.out.println(indent + unit);
        for(Iterator i = unit.getChildren().iterator(); i.hasNext(); ) {
            dumpUnit((TransferUnit)i.next(), indent + "  ");
        }
    }
    
    /**
     * Generates a map of random values for each of
     * the variable definitions of the given protocol.
     */
    private static Map generateValues(Protocol protocol) {
        Map values = new HashMap();
        for(Iterator i = protocol.getVariables().iterator(); i.hasNext(); ) {
            values.put(((VariableDefinition)i.next()).getRawName(), randomString(6, 3));
        }
        return values;
    }
    
    /**
     * Returns a random string.
     */
    private static String randomString(int length, int variance) {
        if(random.nextDouble() < 0.02) {
            return "";
        }
        length = variance + random.nextInt(length);
        StringBuffer sb = new StringBuffer(length);
        for(int i = 0; i < length; i++) {
            sb.append(source.charAt(random.nextInt(source.length())));
        }
        return sb.toString();
    }

    private static String source = "0123456789abcdef";
    private static Random random = new Random();
    
    public static void main(String[] args) {
        // create 8 http units with 2 branches, and dump the result
        getRandomData(8, 2, 2, 0.0f, true);
    }

    
	public static NoteManager getRandomNotes(float totalTime, int numOfNotes, Layer layer) {
		NoteManager noteManager = new NoteManager();
		float step=totalTime/numOfNotes;
		for(int i = 0;i<numOfNotes;i++){
			Note note = new Note(layer, step*i,"Moozoo Note: time="+step*i+", layer="+layer);
            noteManager.addNoteTimeLayer(note);	
		}
		return noteManager;
    }
    
}
