package fi.helsinki.dacopan.model;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import fi.helsinki.dacopan.*;

/**
 * Test case for transfer units.
 * 
 * @version $Id: TransferUnitTestCase.java,v 1.7 2004/05/24 20:25:23 jaarnial Exp $
 */
public class TransferUnitTestCase extends AbstractTestCase {
    
    private TransferUnit unit;
   
    public void testChildren() {
        List children = new ArrayList();
        children.add(createUnit(0.10F, 0.20F));
        children.add(createUnit(0.15F, 0.25F));
        unit = createUnit(children);
        assertEquals("Children were incorrect", children, unit.getChildren());
        for(Iterator i = children.iterator(); i.hasNext(); ) {
            TransferUnit child = (TransferUnit)i.next();
            assertEquals("Child's parent was wrong", unit, child.getParent());
        }
    }

    public void testChildrenAreOrdered() {
        List children = new ArrayList();
        children.add(createUnit(0.2F, 0.5F));
        children.add(createUnit(0.1F, 0.3F));
        List unordered = new ArrayList(children);
        unit = createUnit(children);
        assertEquals(unordered.get(1), unit.getChildren().get(0));
        assertEquals(unordered.get(0), unit.getChildren().get(1));
    }

    public void testChildrenAreNonNull() {
        unit = createUnit(0.01F, 0.1F);
        assertNotNull("Children should be non-null", unit.getChildren());
    }
    
    public void testTimestampsWithNoChildren() {
        unit = createUnit(.01F, .1F);
        assertEquals("Send start incorrect", .01F, unit.getSendStart(), 0);
        assertEquals("Send end incorrect", .01F, unit.getSendEnd(), 0);
        assertEquals("Receive start incorrect", .1F, unit.getReceiveStart(), 0);
        assertEquals("Receive end incorrect", .1F, unit.getReceiveEnd(), 0);
    }
    
    public void testInvalidTimestamps() {
        try {
            unit = createUnit(.1F, .01F);
            fail("Should've failed with illegal arguments");
        } catch(IllegalArgumentException expected) { }
    }
    
    public void testTimestampsWithChildren() {
        List children = new ArrayList();
        children.add(createUnit(.10F, .50F));
        children.add(createUnit(.15F, .70F));
        children.add(createUnit(.20F, .60F));
        unit = createUnit(children);
        assertEquals("Send start incorrect", .10F, unit.getSendStart(), 0);
        assertEquals("Send end incorrect", .20F, unit.getSendEnd(), 0);
        assertEquals("Receive start incorrect", .50F, unit.getReceiveStart(), 0);
        assertEquals("Receive end incorrect", .70F, unit.getReceiveEnd(), 0);
    }
    
    public void testSorting() {
        List unsorted = new ArrayList();
        unsorted.add(createUnit(.050F, .100F));
        unsorted.add(createUnit(.015F, .070F));
        unsorted.add(createUnit(.020F, .060F));
        unsorted.add(createUnit(.050F, .060F));
        unsorted.add(createUnit(.015F, .100F));
        unsorted.add(createUnit(.090F, .100F));
        unsorted.add(createUnit(.070F, .080F));
        unsorted.add(createUnit(.050F, .110F));
        List sorted = new ArrayList(unsorted);
        Collections.sort(sorted);
        assertEquals(unsorted.get(1), sorted.get(0));
        assertEquals(unsorted.get(4), sorted.get(1));
        assertEquals(unsorted.get(2), sorted.get(2));
        assertEquals(unsorted.get(3), sorted.get(3));
        assertEquals(unsorted.get(0), sorted.get(4));
        assertEquals(unsorted.get(7), sorted.get(5));
        assertEquals(unsorted.get(6), sorted.get(6));
        assertEquals(unsorted.get(5), sorted.get(7));
    }
    
    public void testIsOneUnit() {
        TransferUnit a = new TransferUnit(0, Collections.EMPTY_MAP, IPv4, null, HOST_A, HOST_B, 1.0F, 2.0F, false);
        TransferUnit b = new TransferUnit(1, Collections.EMPTY_MAP, IPv4, null, HOST_A, HOST_B, 2.0F, 3.0F, false);
        assertTrue("A is one unit", a.isOneUnit());
        assertTrue("B is one unit", b.isOneUnit());
        TransferUnit c = new TransferUnit(2, Collections.EMPTY_MAP, TCP, null, HOST_A, HOST_B, Arrays.asList(new Object[] { a, b }), false);
        assertFalse("C is not one unit", c.isOneUnit());
    }
    
    public void testGetRoot() {
        TransferUnit a = createUnit(0.1f, 0.2f);
        assertEquals("A's root should be A", a, a.getRoot());
        TransferUnit b = createUnit(0.2f, 0.3f);
        TransferUnit c = createUnit(Arrays.asList(new Object[] { a, b }));
        assertEquals("A's root should be C", c, a.getRoot());
        assertEquals("B's root should be C", c, b.getRoot());
        assertEquals("C's root should be C", c, c.getRoot());
        TransferUnit d = createUnit(Arrays.asList(new Object[] { c }));
        assertEquals("A's root should be D", d, a.getRoot());
        assertEquals("C's root should be D", d, c.getRoot());
    }
    
    public void testHostSpecificVariables() {
        VariableDefinition a = new VariableDefinition("a", VariableDefinition.UNIT, IPv4);
        VariableDefinition b = new VariableDefinition("b", VariableDefinition.UNIT, IPv4);
        TransferUnit unit = createUnit(0.01f, 0.02f);
        unit.addVariableValue(a, null, "A");
        unit.addVariableValue(b, null, "B");
        assertEquals("A", unit.getValue(a));
        assertEquals("B", unit.getValue(b));
        unit = createUnit(0.01f, 0.02f);
        unit.addVariableValue(a, HOST_B, "A");
        unit.addVariableValue(b, HOST_A, "B");
        assertEquals("A", unit.getValue(a));
        assertEquals("B", unit.getValue(b));
        assertNull(unit.getValue(a, HOST_A));
        assertNull(unit.getValue(b, HOST_B));
        assertEquals("A", unit.getValue(a, HOST_B));
        assertEquals("B", unit.getValue(b, HOST_A));
    }

}
